package org.processmining.dataawarereplayer.precision.projection;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XLog;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizerException;
import org.processmining.xesalignmentextension.XAlignmentExtension.MoveType;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentExtensionException;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentMove;

import com.google.common.base.Function;
import com.google.common.collect.ForwardingList;
import com.google.common.collect.Iterators;
import com.google.common.collect.SetMultimap;

public final class ProcessProjectorAlignmentImpl implements ProcessProjector {

	private final static class ProjectedEventImpl implements ProjectedEvent {

		private final XDataAlignmentMove move;
		private final Transition transition;

		private final LogMapping logMapping;

		public ProjectedEventImpl(XAlignmentMove move, Transition transition, LogMapping logMapping) {
			this.move = XDataAlignmentExtension.instance().extendXAlignmentMove(move);
			this.transition = transition;
			this.logMapping = logMapping;
		}

		public Object getAttributeValue(String attributeName) {
			// Contains the process projection values
			XAttribute attribute = move.getEvent().getAttributes().get(attributeName);
			if (attribute != null) {
				return XUtils.getAttributeValue(attribute);
			} else {
				return null;
			}
		}

		public Set<String> getAttributesWritten() {
			XDataAlignmentExtension alignmentExtension = XDataAlignmentExtension.instance();
			Set<String> writtenAttributes = new HashSet<>();
			for (XAttribute attr : move.getEvent().getAttributes().values()) {
				try {
					if (alignmentExtension.isCorrectAttribute(attr) || alignmentExtension.isIncorrectAttribute(attr)
							|| alignmentExtension.isMissingAttribute(attr)) {
						writtenAttributes.add(attr.getKey());
					} else {
						if (logMapping.getExtraAttributes().contains(attr.getKey())) {
							writtenAttributes.add(attr.getKey());
						}
					}
				} catch (XDataAlignmentExtensionException e) {
					throw new IllegalArgumentException("Invalid alignment move: "
							+ XUtils.stringifyEvent(move.getEvent()), e);
				}
			}
			return writtenAttributes;
		}

		public Transition getTransition() {
			return transition;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((getTransition() == null) ? 0 : getTransition().hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			ProjectedEventImpl other = (ProjectedEventImpl) obj;
			if (getTransition() == null) {
				if (other.getTransition() != null)
					return false;
			} else if (!getTransition().equals(other.getTransition()))
				return false;
			return true;
		}

		public String toString() {
			return getTransition().getLabel();
		}

	}

	private static final class ProjectedTraceAlignmentImpl extends ForwardingList<ProjectedEvent> implements
			ProjectedTrace {

		private final List<ProjectedEvent> projectedTrace;
		private final XAlignment alignment;

		public ProjectedTraceAlignmentImpl(XAlignment alignment, LogMapping logMapping) {
			this.projectedTrace = new ArrayList<>(alignment.size());
			this.alignment = alignment;
			for (final XAlignmentMove move : alignment) {
				if (move.getType() != MoveType.LOG) {
					Transition transition = logMapping.getLocalIdToTransition().get(move.getActivityId());
					if (transition == null) {
						throw new IllegalArgumentException(
								"Invalid activity of "
										+ XUtils.stringifyEvent(move.getEvent())
										+ " corresponding Transition with is not know! There should have been a transition with local ID "
										+ move.getActivityId());
					} else {
						projectedTrace.add(new ProjectedEventImpl(move, transition, logMapping));
					}
				}
			}
		}

		protected List<ProjectedEvent> delegate() {
			return projectedTrace;
		}

		public String getName() {
			return alignment.getName();
		}

	}

	private final Iterable<XAlignment> alignedLog;
	private final LogMapping logMapping;

	@SuppressWarnings("rawtypes")
	public ProcessProjectorAlignmentImpl(DataPetriNet net, XLog log, Iterable<XAlignment> alignedLog,
			XEventClasses eventClasses, SetMultimap<String, Transition> eventIdentityToTransition,
			Map<String, String> variableMapping, Map<String, Class> extraAttributes,
			SetMultimap<Transition, String> executors) throws ProcessProjectionException {
		this.alignedLog = alignedLog;
		try {
			logMapping = new LogMapping(net, log, eventClasses, eventIdentityToTransition, variableMapping,
					extraAttributes, executors);
		} catch (StringDiscretizerException e) {
			throw new ProcessProjectionException(e);
		}
	}

	public Iterator<ProjectedTrace> projectedTraces() {
		return Iterators.transform(alignedLog.iterator(), new Function<XAlignment, ProjectedTrace>() {

			public ProjectedTrace apply(XAlignment alignment) {
				return new ProjectedTraceAlignmentImpl(alignment, logMapping);
			}
		});
	}

	@SuppressWarnings("rawtypes")
	public Map<String, Class> getAttributeType() {
		return logMapping.getAttributeType();
	}

	public StringDiscretizer getStringDiscretizer() {
		return logMapping.getDiscretizer();
	}

}