package org.processmining.models.DCNet;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import org.processmining.models.cnet.CNet;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CNetBinding.Type;
import org.processmining.plugins.DataConformance.DCnet.DCNetExecutionTrace;
import org.processmining.plugins.DataConformance.framework.ExecutionTrace;
import org.processmining.plugins.DataConformance.framework.LogExecutionTrace;
import org.processmining.plugins.DataConformance.framework.Replayable;

abstract class Domain { }

class DiscreteDomain extends Domain
{
	private final HashSet<Comparable> values=new HashSet<Comparable>();

	public Set<Comparable> getValues() {
		return values;
	}
	
	
}

@SuppressWarnings("unchecked")
class ContinuousDomain extends Domain
{
	public ContinuousDomain(Class domainType, Comparable minValue, Comparable maxValue) {
		this.domainType = domainType;
		this.minValue = minValue;
		this.maxValue = maxValue;
	}
	
	private final Class domainType;
	private final Comparable minValue;
	private final Comparable maxValue;
	public Class getDomainType() {
		return domainType;
	}
	public Comparable getMinValue() {
		return minValue;
	}
	public Comparable getMaxValue() {
		return maxValue;
	}
	
	
}

@SuppressWarnings("unchecked")
public class DCNet extends CNet implements Replayable 
{
	public enum Position 
	{
		BEFORE, AFTER
	}


	public static final String SPECIAL_CHAR = "^!+%/*-&|<=>(), ";	
	
	public static char OUTPUTVAR_IDENTIFIER='$';
	public static Position OUTPUTVAR_IDENTIFIER_POS = Position.BEFORE;
	
	public static Set<String> getWrittenVariables(String condition) 
	{
		StringTokenizer tokenizer=new StringTokenizer(condition, DCNet.SPECIAL_CHAR);
		HashSet<String> writtenVariables=new HashSet<String>();
		while(tokenizer.hasMoreTokens())
		{
			String nextVar=tokenizer.nextToken();
			if (nextVar.length()>0)
			{
				if (DCNet.OUTPUTVAR_IDENTIFIER_POS==DCNet.Position.AFTER && nextVar.endsWith(""+DCNet.OUTPUTVAR_IDENTIFIER))
					writtenVariables.add(nextVar.substring(0,nextVar.length()));
				else if (DCNet.OUTPUTVAR_IDENTIFIER_POS==DCNet.Position.BEFORE && nextVar.startsWith(""+DCNet.OUTPUTVAR_IDENTIFIER))
					writtenVariables.add(nextVar.substring(1));
			}
		}
		return writtenVariables;
	}
	protected Map<String, Domain> setTable=new HashMap<String,Domain>();
	protected Map<String, String> variableTable=new HashMap<String,String>();  
	protected Map<DCNetNode, Set<String>> lastNodesPerVariables=new HashMap<DCNetNode, Set<String>> ();
	private static String RESOURCE_NAME="ResourceName"; 
	
	private String resourceName=null;
	

	protected int shortestTrace=0;	
	
	public DCNet(String label) 
	{
		super(label);
	}
	
	public CNetBinding addInputBinding(CNetNode node, CNetNode... nodes) 
	{
		return this.addInputBinding((DCNetNode)node, "", (DCNetNode[])nodes); 
	}	
	
	public CNetBinding addInputBinding(CNetNode node, Collection<? extends CNetNode> nodes) 
	{
		return this.addInputBinding((DCNetNode)node, "", (Collection<? extends DCNetNode>)nodes);  
	}

	/**
	 * Add an input binding
	 * 
	 * @param node
	 */
	public DCNetBinding addInputBinding(DCNetNode node, String dataExpression, Collection<? extends DCNetNode> nodes) {
		DCNetBinding binding = new DCNetBinding(Type.INPUT, node, nodes,  dataExpression);
		inputBindings.get(node).add(binding);
		return binding;
	}	
	
	/**
	 * Add an input binding
	 * 
	 * @param node
	 */
	public DCNetBinding addInputBinding(DCNetNode node, String dataExpression, DCNetNode... nodes) {
		DCNetBinding binding = new DCNetBinding(Type.INPUT, node, nodes,  dataExpression);
		inputBindings.get(node).add(binding);
		return binding;
	}
	
	public CNetNode addNode(DCNetNode node, String resourceSet) {
		if (getSetValues(resourceSet)==null)
			throw new IllegalStateException("The set "+resourceSet+" is not defined");
		if (!getSetValues(getSetForVariable(resourceName)).containsAll(getSetValues(resourceSet)))
			throw new IllegalStateException("The set "+resourceSet+" is incompatible with the resource set");
		node.setExecutor(resourceName,resourceSet);
		CNetNode retValue=this.addNode(node);
		this.addNonAffectingVariables(resourceName, node);
		return retValue;
	}
	
	public void addNonAffectingVariables(String varName, DCNetNode node)
	{
		if (!variableTable.containsKey(varName) || !this.getNodes().contains(node))
			throw new IllegalArgumentException("Variable "+varName+" or node "+node+" does not exist");
		Set<String> varNameSet=lastNodesPerVariables.get(node);
		if (varNameSet==null) {
			varNameSet=new HashSet<String>();
			lastNodesPerVariables.put(node,varNameSet);
		}
		varNameSet.add(varName);
	}
	
	
	public CNetBinding addOutputBinding(CNetNode node, CNetNode... nodes) 
	{
		return this.addOutputBinding((DCNetNode)node, "", (DCNetNode[])nodes); 
	}
	
	public CNetBinding addOutputBinding(CNetNode node, Collection<? extends CNetNode> nodes) 
	{
		return this.addOutputBinding((DCNetNode)node, "", (Collection<? extends DCNetNode>)nodes);
	}

	/**
	 * Add an output binding
	 * 
	 * @param node
	 */
	public DCNetBinding addOutputBinding(DCNetNode node, String dataExpression, Collection<? extends DCNetNode> nodes) 
	{
		DCNetBinding binding = new DCNetBinding(Type.OUTPUT, node, nodes, dataExpression);
		outputBindings.get(node).add(binding);
		return binding;
	}

	/**
	 * Add an output binding
	 * 
	 * @param node
	 */
	public DCNetBinding addOutputBinding(DCNetNode node, String dataExpression, DCNetNode... nodes) {
		DCNetBinding binding = new DCNetBinding(Type.OUTPUT, node, nodes,  dataExpression);
		outputBindings.get(node).add(binding);
		return binding;
	}

	public void setDomain(String setName, Class type, Comparable minValue, Comparable maxValue)
	{
		Domain valueSet=setTable.get(setName);
		if (valueSet!=null)
			new IllegalArgumentException("The set is already defined over a discrete domain");
		if (valueSet instanceof DiscreteDomain)
			new IllegalArgumentException("The set is already defined over a discrete domain");
		if (type != minValue.getClass() || type != maxValue.getClass())
			new IllegalArgumentException("The set is already defined over a discrete domain");
		new ContinuousDomain(type, minValue, maxValue);
	}

	public void addValuesToSet(String setName, Comparable... newValues)
	{
		Domain valueSet=setTable.get(setName);
		if (valueSet==null)
		{
			valueSet=new DiscreteDomain();
			setTable.put(setName, valueSet);
		}
		if (valueSet instanceof ContinuousDomain)
			new IllegalArgumentException("The set is already defined over a continuous domain");
		Set<Comparable> valuesDomain=((DiscreteDomain) valueSet).getValues();
		for(Comparable newValue : newValues)
			valuesDomain.add(newValue);
	}

	public void addVariable(String varName, String setName) 
	{
		if (new StringTokenizer(varName, SPECIAL_CHAR+OUTPUTVAR_IDENTIFIER).countTokens()>1)
			new IllegalArgumentException("The variable "+varName+ " contains illegal characters");
		variableTable.put(varName, setName);	
	}	

	public boolean areNonAffectingVariables(DCNetNode node) {
		return (lastNodesPerVariables.get(node)!=null);
	}

	public ExecutionTrace buildEmptyPrefix() 
	{
		DCNetExecutionTrace emptyTrace=new DCNetExecutionTrace(this);
		return emptyTrace;
	}

	public ExecutionTrace createLogExecutionTrace(ExecutionTrace sigmaT) {
		return new LogExecutionTrace(sigmaT);
	}
	
	

	public Set<String> getNonAffectingVariables(DCNetNode node)
	{
		Set<String> varNameSet=lastNodesPerVariables.get(node);
		if (varNameSet==null) {
			varNameSet=new HashSet<String>();
		}
		return Collections.unmodifiableSet(varNameSet);
	}	
	
	/**
	 * Returns the predecessors of a CNet node, i.e. each of these nodes is in
	 * an input binding of the given node. If the net is consistent, then the
	 * given node is also in at least one output binding of each node in the
	 * returned set.
	 * 
	 * @param node
	 * @return
	 */
	public Set<CNetNode> getPredecessors(CNetNode node) 
	{
		if (node==this.getStartNode())
			return new HashSet<CNetNode>();
		else
			return super.getPredecessors(node);
	}	
	
	public String getSetForVariable(String varName)
	{
		return variableTable.get(varName);
	}

	public Map<String, Set<Comparable>> getSetTable() {
		HashMap<String, Set<Comparable>> retValue=new HashMap<String, Set<Comparable>>();
		for(String set : setTable.keySet())
			retValue.put(set, getSetValues(set));
		return retValue;
	}
	
	
	public Set<Comparable> getSetValues(String setName)
	{
		Domain aSet=setTable.get(setName);
		if (aSet == null)
			return null;
		if (aSet instanceof DiscreteDomain)
			return ((DiscreteDomain)aSet).getValues();
		else {
			ContinuousDomain domain=(ContinuousDomain)aSet;
			HashSet<Comparable> values=new HashSet<Comparable>();
			values.add(domain.getMinValue());
			values.add(domain.getMaxValue());
			return values;
		}
	}
	
	public Set<CNetNode> getSuccessors(CNetNode node) 
	{
		if (node==this.getEndNode())
			return new HashSet<CNetNode>();
		else
			return super.getSuccessors(node);
	}
	
	public Set<String> getVariables()
	{
		return Collections.unmodifiableSet(variableTable.keySet());
	}

	/**
	 * check if this cnet is consistent with the formal definitions:
	 * 
	 * - one start node
	 * 
	 * - one end node
	 * 
	 * - dependencies between nodes are reflected in input/output bindings
	 * 
	 * * @return
	 */
	@Override
	public boolean isConsistent() {
		if (endNode == null) {
			return false;
		} 
		if (startNode == null) {
			return false;
		} 
		for (CNetNode node : nodes) {
			if (!(node instanceof DCNetNode))
				return false;
			DCNetNode dnode=(DCNetNode) node;
			if (getInputBindings(dnode).isEmpty() && !dnode.equals(startNode)) {
				return false;
			}
			if (getOutputBindings(dnode).isEmpty() && !dnode.equals(endNode)) {
				return false;
			}
			if (!dnode.equals(startNode))
				for (CNetBinding binding : getInputBindings(dnode)) {
					nn: for (CNetNode n2 : binding.getBoundNodes()) {
						for (CNetBinding b2 : getOutputBindings(n2)) {
							if (b2.getBoundNodes().contains(dnode)) {
								continue nn;
							}
						}
						// No binding found 
						return false;
					}
				}
			if (!dnode.equals(endNode))
				for (CNetBinding binding : getOutputBindings(dnode)) {
					nn: for (CNetNode n2 : binding.getBoundNodes()) {
						for (CNetBinding b2 : getInputBindings(n2)) {
							if (b2.getBoundNodes().contains(dnode)) {
								continue nn;
							}
						}
						// No binding found 
						return false;
					}
				}
		}

		return true;
	}

	public int lengthShortestTrace() {
		return shortestTrace;
	}

	public void setEndNode(CNetNode endNode) 
	{
		super.setEndNode(endNode);
		addOutputBinding((DCNetNode)endNode,"",DCNetNode.emptyNode);
	}

	public void setLengthShortestTrace(int shortestTrace) {
		this.shortestTrace = shortestTrace;
	}

	public void setResourceSet(String setName) {
		addVariable(RESOURCE_NAME, setName);
		resourceName=RESOURCE_NAME;
	}

	public void setStartNode(CNetNode startNode) 
	{
		super.setStartNode(startNode);
		addInputBinding((DCNetNode)startNode,"",DCNetNode.emptyNode);
	}
}
