package org.processmining.plugins.DataConformance.framework;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.DataConformance.Utility;

public class ReplayingHelper 
{
	XTrace trace; 
	Map<ReplayableActivity, XEventClass> activityMapping; 
	Map<String, String> variableMapping; 
	XEventClasses eventClasses;
	Replayable model;
	ActivityMatchCosts activityMatchCosts; 
	VariableMatchCosts variableMatchCosts;	
	Map<String, Float> maxDistance;
	float minCost;

	public ReplayingHelper(Replayable model, XTrace trace, Map<ReplayableActivity, XEventClass> activityMapping,
			Map<String, String> variableMapping, XEventClasses eventClasses, Map<String, Float> maxDistance, ActivityMatchCosts activityCosts,
			VariableMatchCosts variableCosts) 
	{
		this.model=model;
		this.trace=trace;
		this.activityMapping=activityMapping;
		this.variableMapping=variableMapping;
		this.eventClasses=eventClasses;
		this.activityMatchCosts=activityCosts;
		this.variableMatchCosts=variableCosts;
		this.maxDistance=maxDistance;
		this.minCost=Math.min(activityMatchCosts.lowestCost(), variableMatchCosts.lowestCost());
	}

	public ExecutionTrace buildLogSequence() 
	{
		XAttribute attr=trace.getAttributes().get("concept:name");
		String traceName="";
		if (attr!=null)
			traceName=((XAttributeLiteral)attr).getValue();		
		LogExecutionTrace retValue=new LogExecutionTrace(traceName);
		for (XEvent event : trace)
		{
			ExecutionStep step=getExecutionStep(event);
			if (step!=null)
				retValue.add(step);
		}
		return retValue;
	}
	
	public Collection<ReplayState> createSuccessors(ReplayState mu) 
	{
		List<ReplayState> successors=new LinkedList<ReplayState>();
		//It always includes the prefix extended with a BOTTOM-step.
		Collection<ExecutionTrace> extendedLogPrefixes=mu.logTracePrefix.successors(null);
		for(ExecutionTrace extendedLogPrefix : extendedLogPrefixes)
		{	
			ExecutionStep lastLogStep=extendedLogPrefix.getLast();
			Collection<ExecutionTrace> extendedProcessPrefixes;
			extendedProcessPrefixes=mu.processTracePrefix.successors(lastLogStep);
	
			for(ExecutionTrace extendedProcessPrefix : extendedProcessPrefixes)
			{
				//Move in the log + Move in the process
				//If lastLogStep==ExecutionStep.tauStep, then it's Not move in the log + Move in the process
				ReplayState state = new ReplayState();
				state.processTracePrefix = extendedProcessPrefix;
				state.logTracePrefix = extendedLogPrefix;
				successors.add(state);
			}

			
			if(lastLogStep!=ExecutionStep.bottomStep)
			{
				//Move in the log + Not move in the process
				ReplayState state = new ReplayState();
				state.logTracePrefix = (ExecutionTrace) extendedLogPrefix.clone();
				state.processTracePrefix = (ExecutionTrace) mu.processTracePrefix.clone();
				state.processTracePrefix.add(ExecutionStep.bottomStep);
				successors.add(state);
			}
		}
		return successors;
	}

	public float computeHeuristicValue(ReplayState state, ExecutionTrace completeLogTrace) {

		int logEventToReplay=completeLogTrace.size()-state.logTracePrefix.lengthExecutionTraceWithoutBottomSteps();
		return minCost*logEventToReplay;
	}
	
	public float computeDistance(ReplayState prevState, ReplayState nextState) {
		float lastStepCost = costMatchingLastStep(nextState);
		if (!nextState.logTracePrefix.getLast().equals(ExecutionStep.bottomStep))
			lastStepCost+=minCost;
		else
			if (model.lengthShortestTrace() > 0)
				if (nextState.processTracePrefix.lengthExecutionTraceWithoutBottomSteps() <= model.lengthShortestTrace()) {
					lastStepCost+=minCost;
				}
		return lastStepCost;
	}

	private float costMatchingLastStep(ReplayState state) 
	{
		float cost=0;
		ExecutionStep lastLogStep=state.logTracePrefix.getLast();
		ExecutionStep lastProcessStep=state.processTracePrefix.getLast();
	
		if (lastLogStep==ExecutionStep.bottomStep)
			return activityMatchCosts.costMoveOnlyInProcess(lastProcessStep.getActivity());
		else if (lastProcessStep==ExecutionStep.bottomStep)
			return activityMatchCosts.costMoveOnlyInLog(lastLogStep.getActivity());
		
		if (lastLogStep.getActivity().equals(lastProcessStep.getActivity())) {
			String activity=lastLogStep.getActivity();
			//For each variable var written in the log step...
			for(String var : lastLogStep.keySet()) {
				Object logValue=lastLogStep.get(var);
				Object processValue=lastProcessStep.get(var);
				
				//If var has also been written in the process step, compute the distance between the values 
				if (processValue!=null)	{
					if (processValue instanceof String) {
						String logValueAsString=String.valueOf(logValue);
						float distance=org.processmining.plugins.DataConformance.Utility.stringDistance((String) processValue,logValueAsString);
						if (distance>0)
						cost+=variableMatchCosts.costFaultyValue(activity,var);//maxDistance.get(var)*distance;
					}
					else if (processValue instanceof Number) 
					{
						float logValueAsFloat;
						if (logValue instanceof String)
							logValueAsFloat=Float.parseFloat((String) logValue);
						else
							logValueAsFloat=((Number)logValue).floatValue();
						float distance=Math.abs(((Number)processValue).floatValue()-logValueAsFloat);
						if (distance>0)
						cost+=variableMatchCosts.costFaultyValue(activity,var);//maxDistance.get(var)*distance;					
					}
					else if(!processValue.equals(logValue))
					{
						cost+=variableMatchCosts.costFaultyValue(activity, var);
					}
				}		
				//If var has not been written, then 
				else 
					cost+=variableMatchCosts.notWritingCost(activity,var);	
			}
			
			for(String var : lastProcessStep.keySet())
				if (lastLogStep.get(var)==null)
					cost+=variableMatchCosts.notWritingCost(activity,var);	
		}
		return cost;
	}

	protected ExecutionStep getExecutionStep(XEvent event)
	{
		String stepName=null;
		for(Entry<ReplayableActivity, XEventClass>  activityMappingEntry : activityMapping.entrySet()) {
			if (activityMappingEntry.getValue().equals(eventClasses.getClassOf(event))) {
				stepName=activityMappingEntry.getKey().getLabel();
				break;
			}
				
		}
		if (stepName==null) {
			//stepName=(String) Utility.getValue(event.getAttributes().get("concept:name"));
			return null;
		}
		ExecutionStep step=new ExecutionStep(stepName);
		XAttributeMap attributes=event.getAttributes();
		for(Entry<String, XAttribute> attribute : attributes.entrySet())
		{ 
			Object value=Utility.getValue(attribute.getValue());
			if (value!=null)
			{
				for(Entry<String, String> entr : variableMapping.entrySet()) {
					if (attribute.getKey().equals(entr.getValue())) {
						step.put(entr.getKey(),value);
						break;
					}
				}
			}
		}
		return step;
	}
}
