/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import nl.tue.astar.AStarThread;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.petrinet.replayresult.StepTypes;

import com.google.common.collect.AbstractIterator;

/**
 * @author F. Mannhardt
 * 
 */
public final class ControlFlowAlignmentResultImpl implements ControlFlowAlignmentResult {

	//TODO support reverse iteration
	private final class CostRecordingIterator extends AbstractIterator<ControlFlowAlignmentResult.Step> {

		//		private final Iterator<XEvent> eventIterator;
		private final boolean[] usedEvents;
		private final Iterator<DRecord> recordIterator;
		// Current record in recordIterator
		private DRecord currentRec;

		private boolean checkForMoreUnmappedEvents = false;

		private float mmCost;
		private float mlCost;
		private float mSyncCost;
		private double delta;

		private CostRecordingIterator() {
			//			this.eventIterator = trace.iterator();
			this.usedEvents = new boolean[trace.size()];
			this.recordIterator = Arrays.asList(record.getHistory()).iterator();
			this.delta = delegate.getDelta();
		}

		public float getCostSoFar() {
			return mmCost + mlCost + mSyncCost;
		}

		protected Step computeNext() {

			if (checkForMoreUnmappedEvents) {
				//				assert eventIterator.hasNext() : "Could not find matching event in trace for record" + currentRec
				//						+ " Error in Alignment!";
				short activityId = (short) linearTrace.get(currentRec.getMovedEvent()); // a is the event obtained from the replay
				usedEvents[currentRec.getMovedEvent()] = true;
				XEvent event = trace.get(currentRec.getMovedEvent());// eventIterator.next();
				XEventClass eventClass = mapEv2EventClass.get(event); // this is the current event
				if (delegate.getIndexOf(eventClass) != activityId) {
					return returnUnmappedEvent(eventClass, event);
				} else {
					// Found the correct event stop looking for more 
					checkForMoreUnmappedEvents = false;
					if (currentRec.getModelMove() == AStarThread.NOMOVE) {
						return returnLogMove(currentRec, activityId, event);
					} else {
						return returnSyncMove(currentRec, event);
					}
				}
			} else {
				if (recordIterator.hasNext()) {
					currentRec = recordIterator.next();

					if (currentRec.getMovedEvent() == AStarThread.NOMOVE) {
						return returnModelMove(currentRec, delta);
					}

					// a move occurred in the log. Check if class aligns with class in trace
					short activity = (short) linearTrace.get(currentRec.getMovedEvent()); // a is the event obtained from the replay
					//					XEvent event = eventIterator.next();
					usedEvents[currentRec.getMovedEvent()] = true;
					XEvent event = trace.get(currentRec.getMovedEvent());
					XEventClass clsInTrace = mapEv2EventClass.get(event); // this is the current event
					if (delegate.getIndexOf(clsInTrace) != activity) {
						// The next event in the trace is not of the same class as the next event in the A-star result.
						// This is caused by the class in the trace not being mapped to any transition.
						// move log only
						checkForMoreUnmappedEvents = true; // Remember that there might be more
						return returnUnmappedEvent(clsInTrace, event);
					}
					if (currentRec.getModelMove() == AStarThread.NOMOVE) {
						return returnLogMove(currentRec, activity, event);
					} else {
						return returnSyncMove(currentRec, event);
					}
				} else {
					if (!isPartialResult) {
						for (int i = 0; i < usedEvents.length; i++) {
							if (!usedEvents[i]) {
								usedEvents[i] = true;
								// add the rest of the trace
								XEvent event = trace.get(i);
								XEventClass eventClass = mapEv2EventClass.get(event);
								return returnUnmappedEvent(eventClass, event);
							}
						}
						return returnEndOfData();
					} else {
						return returnEndOfData();
					}
				}
			}
		}

		private Step returnEndOfData() {
			// Performance optimization, we don't need a second iteration
			finalTraceCost = getCostSoFar();
			return endOfData();
		}

		private Step returnModelMove(DRecord currentRec, double delta) {
			StepTypes type;
			// move model only
			Transition transition = delegate.getTransition((short) currentRec.getModelMove());
			if (transition.isInvisible()) {
				type = StepTypes.MINVI;
			} else {
				type = StepTypes.MREAL;
			}
			mmCost += (delegate.getCostForMoveModel((short) currentRec.getModelMove()) - delegate.getEpsilon()) / delta;
			return new StepImpl(currentRec, transition, null, type);
		}

		private Step returnSyncMove(DRecord currentRec, XEvent event) {
			Object node = delegate.getTransition((short) currentRec.getModelMove());
			mSyncCost += (delegate.getCostForMoveSync((short) currentRec.getModelMove()) - delegate.getEpsilon())
					/ delta;
			return new StepImpl(currentRec, (Transition) node, event, StepTypes.LMGOOD);
		}

		private Step returnLogMove(DRecord currentRec, short activity, XEvent event) {
			mlCost += (delegate.getCostForMoveLog(activity) - delegate.getEpsilon()) / delta;
			return new StepImpl(currentRec, null, event, StepTypes.L);
		}

		private Step returnUnmappedEvent(XEventClass unmappedEventClass, XEvent event) {
			mlCost += mapEvClass2Cost.get(unmappedEventClass);
			return new StepImpl(null, null, event, StepTypes.L);
		}

	}

	private final static class StepImpl implements Step {

		private final XEvent logNode;
		private final Transition modelNode;
		private final StepTypes type;
		private final DRecord record;

		public StepImpl(DRecord record, Transition modelNode, XEvent logNode, StepTypes type) {
			this.record = record;
			this.modelNode = modelNode;
			this.logNode = logNode;
			this.type = type;
		}

		public StepTypes getType() {
			return type;
		}

		public Transition getModelNode() {
			return modelNode;
		}

		public XEvent getLogNode() {
			return logNode;
		}

		public DRecord getRecordForStep() {
			return record;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((logNode == null) ? 0 : logNode.hashCode());
			result = prime * result + ((modelNode == null) ? 0 : modelNode.hashCode());
			result = prime * result + ((type == null) ? 0 : type.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof StepImpl))
				return false;
			StepImpl other = (StepImpl) obj;
			if (logNode == null) {
				if (other.logNode != null)
					return false;
			} else if (!logNode.equals(other.logNode))
				return false;
			if (modelNode == null) {
				if (other.modelNode != null)
					return false;
			} else if (!modelNode.equals(other.modelNode))
				return false;
			if (type != other.type)
				return false;
			return true;
		}

		public String toString() {
			switch (type) {
				case L :
					return String.format("%s (%s)", XUtils.getConceptName(logNode), type);
				case LMSWAPPED :
				case LMREPLACED :
				case LMNOGOOD :
				case LMGOOD :
				case MINVI :
				case MREAL :
				default :
					return String.format("%s (%s)", modelNode.getLabel(), type);
			}
		}

	}

	private final DRecord record;

	private final Map<XEvent, XEventClass> mapEv2EventClass;
	private final Map<XEventClass, Integer> mapEvClass2Cost;
	private final AbstractPDelegate<? extends Tail> delegate;

	private final XTrace trace;
	private final List<XEvent> mappedTrace;
	private final Trace linearTrace;

	private final boolean isPartialResult;
	private float finalTraceCost;

	public ControlFlowAlignmentResultImpl(DRecord record, AbstractPDelegate<? extends Tail> delegate,
			Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost, XTrace trace,
			List<XEvent> mappedTrace, Trace linearTrace) {
		this(record, delegate, mapEv2EventClass, mapEvClass2Cost, trace, mappedTrace, linearTrace, false);
	}

	public ControlFlowAlignmentResultImpl(DRecord record, AbstractPDelegate<? extends Tail> delegate,
			Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost, XTrace trace,
			List<XEvent> mappedTrace, Trace linearTrace, boolean isPartialResult) {
		this.record = record;
		this.delegate = delegate;
		this.mapEv2EventClass = mapEv2EventClass;
		this.mapEvClass2Cost = mapEvClass2Cost;
		this.trace = trace;
		this.mappedTrace = mappedTrace;
		this.linearTrace = linearTrace;
		this.isPartialResult = isPartialResult;
	}

	public DRecord getLastRecord() {
		return record;
	}

	public float getControlFlowCost() {
		if (finalTraceCost > -1) {
			return finalTraceCost;
		} else {
			// Not yet iterated, we need to 
			CostRecordingIterator costRecordingIterator = new CostRecordingIterator();
			while (costRecordingIterator.hasNext()) {
				costRecordingIterator.next();
			}
			assert finalTraceCost > -1 : "Error in cost calculation";
			return finalTraceCost;
		}
	}

	public String toString() {
		return buildAlignmentString() + " Costs: " + getControlFlowCost();
	}

	private String buildAlignmentString() {
		StringBuilder sb = new StringBuilder("(");
		Iterator<Step> stepIter = getSteps().iterator();
		while (stepIter.hasNext()) {
			Step step = stepIter.next();
			if (step.getModelNode() != null) {
				sb.append(step.getModelNode());
			} else {
				sb.append(XUtils.getConceptName(step.getLogNode()));
			}
			sb.append("[");
			sb.append(step.getType());
			sb.append("]");
			if (stepIter.hasNext()) {
				sb.append(',');
			}
		}
		sb.append(')');
		return sb.toString();
	}

	public int getApproximateSize() {
		return record.getBacktraceSize() + 1;
	}

	public XEvent getLastEvent() {
		int eventMove = record.getMovedEvent();
		if (eventMove != AbstractAStarThread.NOMOVE) {
			return mappedTrace.get(eventMove);
		} else {
			return null;
		}
	}

	public Transition getLastTransition() {
		int modelMove = record.getModelMove();
		if (modelMove != AbstractAStarThread.NOMOVE) {
			return delegate.getTransition((short) modelMove);
		} else {
			return null;
		}
	}

	public Iterable<Step> getSteps() {
		return new Iterable<ControlFlowAlignmentResult.Step>() {

			public Iterator<Step> iterator() {
				return new CostRecordingIterator();
			}
		};
	}

}