/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow.adapter;

import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.Hash.Strategy;
import it.unimi.dsi.fastutil.objects.Object2ShortMap;
import it.unimi.dsi.fastutil.objects.Object2ShortOpenHashMap;
import it.unimi.dsi.fastutil.shorts.Short2ShortMap;
import it.unimi.dsi.fastutil.shorts.Short2ShortOpenHashMap;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import nl.tue.astar.AStarObserver;
import nl.tue.astar.AStarThread.ASynchronousMoveSorting;
import nl.tue.astar.AStarThread.Canceller;
import nl.tue.astar.AStarThread.QueueingModel;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread.StorageHandler;
import nl.tue.astar.util.LinearTrace;
import nl.tue.astar.util.PartiallyOrderedTrace;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.AbstractPILPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.controlflow.resultbuilder.AlignmentResultBuilder;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapterWithoutPartialImpl;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserver;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;
import org.processmining.plugins.balancedconformance.visualisation.DotDataAStarObserver;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import com.google.common.base.Joiner;
import com.google.common.collect.ImmutableMap;

/**
 * Base class for integrated data & control-flow alignment
 * 
 * @author F. Mannhardt
 * 
 */
public abstract class AlignmentAbstractAdapter<T extends Tail> implements AlignmentAdapter {

	public interface EmptyTraceResult {
		DataAlignedTrace getDataAlignment();

		ControlFlowAlignmentResult getControlFlowAlignment();
	}

	public static class ControlFlowAlignmentConfig {
		public Canceller canceller = new Canceller() {

			public boolean isCancelled() {
				return Thread.interrupted();
			}
		};

		public Marking initialMarking;
		public Marking[] finalMarkings;

		public Map<Transition, Integer> mapTransition2Cost;
		public Map<XEventClass, Integer> mapEvClass2Cost;
		public TransEvClassMapping transitionMapping;

		public int maxNumOfStates = Integer.MAX_VALUE;
		public double timeLimit = Double.MAX_VALUE;
		public boolean activateCache = true;
		public boolean useOptimizations = true;
		public ASynchronousMoveSorting sorting = ASynchronousMoveSorting.LOGMOVEFIRST;
		public QueueingModel queueingModel = QueueingModel.DEPTHFIRST;

		public int costDelta;

		public boolean usePartialOrder;

	}

	protected static boolean useObserver = false;

	protected final Map<Transition, Integer> mapTrans2Cost;
	protected final Map<XEventClass, Integer> mapEvClass2Cost;
	protected int maxNumOfStates;
	protected double timeLimit;
	protected final Marking initMarking;
	protected final Marking[] finalMarkings;
	protected final DataPetriNet net;
	protected final XEventClasses classes;
	protected final TransEvClassMapping mapping;

	protected final Canceller canceller;
	protected final QueueingModel queueingModel;
	protected final ASynchronousMoveSorting sorting;
	protected final AbstractPDelegate<T> delegate;

	protected final DataConformancePlusObserver dataConformancePlusObserver;
	protected final DataAlignmentAdapter dataAdapter;
	protected final DataViewBuilder dataViewBuilder;
	protected final LogMapping logMapping;

	protected final boolean activateCache;
	protected final boolean useOptimizations;

	protected final StorageHandlerFactory<T> storageFactory;

	protected StorageHandler<PHead, T> storageHandler;
	protected DataStateStore dataStateStore;

	public AlignmentAbstractAdapter(XLog log, DataPetriNet net, LogMapping logMapping, ControlFlowAlignmentConfig c,
			StorageHandlerFactory<T> storageFactory, DataConformancePlusObserver dataConformancePlusObserver,
			DataAlignmentAdapter dataAdapter) {
		this(logMapping, net, c.initialMarking, c.finalMarkings, c.costDelta, c.mapTransition2Cost, log,
				c.mapEvClass2Cost, c.transitionMapping, c.maxNumOfStates, c.timeLimit, c.canceller,
				dataConformancePlusObserver, dataAdapter, c.activateCache, c.useOptimizations, c.sorting,
				c.queueingModel, storageFactory);
	}

	private AlignmentAbstractAdapter(LogMapping logMapping, DataPetriNet net, Marking initialMarking,
			Marking[] finalMarkings, int delta, Map<Transition, Integer> mapTrans2Cost, XLog log,
			Map<XEventClass, Integer> mapEvClass2Cost, final TransEvClassMapping mapping, int maxNumOfStates,
			double timeLimit, Canceller canceller, DataConformancePlusObserver dataConformancePlusObserver,
			DataAlignmentAdapter dataAdapter, boolean activateCache, boolean useOptimizations,
			ASynchronousMoveSorting sorting, QueueingModel queueingModel, StorageHandlerFactory<T> storageFactory) {
		super();
		this.logMapping = logMapping;
		this.net = net;
		this.initMarking = initialMarking;
		this.finalMarkings = finalMarkings;
		this.mapTrans2Cost = mapTrans2Cost;
		this.mapEvClass2Cost = mapEvClass2Cost;
		this.mapping = mapping;
		this.maxNumOfStates = maxNumOfStates;
		this.timeLimit = timeLimit;
		this.canceller = canceller;
		this.dataConformancePlusObserver = dataConformancePlusObserver;
		this.dataAdapter = dataAdapter;
		this.activateCache = activateCache;
		this.useOptimizations = useOptimizations;
		this.queueingModel = queueingModel;
		this.storageFactory = storageFactory;
		// for sake of compatibility, insert cost of dummy xevent class
		this.mapEvClass2Cost.put(mapping.getDummyEventClass(), 0);
		this.classes = logMapping.getEventClasses();
		this.sorting = sorting;
		this.delegate = getPDelegate(net, log, classes, mapping, mapTrans2Cost, mapEvClass2Cost, delta, false,
				finalMarkings);

		this.storageHandler = storageFactory.newStorageHandler(delegate);
		this.dataStateStore = storageFactory.newDataStateStore();

		this.dataViewBuilder = new DataViewBuilder(net, delegate, classes, logMapping);
	}

	protected abstract AlignmentResultBuilder createResultBuilder(TraceMapping traceMapping,
			Map<XEvent, XEventClass> emptyMap, Trace linearTrace, AStarObserver observer);

	@Override
	public AlignmentResultBuilder calcReplayResult(TraceMapping traceMapping, boolean usePartialOrder) throws ControlFlowAlignmentException {
		XTrace trace = traceMapping.getTrace();

		Trace  linearTrace;
		if (usePartialOrder){
			linearTrace = convertToPartialOrderedTrace(delegate, traceMapping);
		} else {
			linearTrace = convertToLinearTrace(delegate, traceMapping);
		}

		Map<XEvent, XEventClass> mapEv2EventClass = buildEventClassMap(trace);

		PartialResultFactory partialResultFactory = new PartialResultFactory(delegate, mapEv2EventClass,
				mapEvClass2Cost, trace, traceMapping.getMappedTrace(), linearTrace);
		traceMapping.setPartialResultFactory(partialResultFactory);

		AStarObserver observer = null;
		if (useObserver) {
			String traceName = traceMapping.getTraceName();
			File outputFile = new File((traceName != null ? traceName : UUID.randomUUID().toString()) + ".dot");
			observer = new DotDataAStarObserver(outputFile, delegate, mapEv2EventClass, mapEvClass2Cost, traceMapping,
					linearTrace, getStateGraphStrategy(traceMapping, dataAdapter));
		}
		return createResultBuilder(traceMapping, mapEv2EventClass, linearTrace, observer);
	}

	public EmptyTraceResult calcEmptyTraceAlignment() throws ControlFlowAlignmentException {
		// create an empty log 
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog dummyLog = factory.createLog();
		XTrace emptyTrace = factory.createTrace();
		Trace linearTrace = new LinearTrace("Empty", 0);
		dummyLog.add(emptyTrace);

		try {
			TraceMapping traceMapping = new TraceMapping(emptyTrace, logMapping);

			PartialResultFactory partialResultFactory = new PartialResultFactory(delegate,
					ImmutableMap.<XEvent, XEventClass>of(), mapEvClass2Cost, emptyTrace, emptyTrace, linearTrace);
			traceMapping.setPartialResultFactory(partialResultFactory);

			AStarObserver observer = null;
			if (useObserver) {
				String traceName = traceMapping.getTraceName();
				File outputFile = new File((traceName != null ? traceName : UUID.randomUUID().toString()) + ".dot");
				observer = new DotDataAStarObserver(outputFile, delegate, Collections.<XEvent, XEventClass>emptyMap(),
						mapEvClass2Cost, traceMapping, linearTrace, getStateGraphStrategy(traceMapping, dataAdapter));
			}

			AlignmentResultBuilder resultBuilder = createResultBuilder(traceMapping,
					Collections.<XEvent, XEventClass>emptyMap(), linearTrace, observer);
			final ControlFlowAlignmentResult cfAlignment = resultBuilder.calcResult(Double.MAX_VALUE);
			final DataAlignedTrace dataAlignment = dataAdapter.retrieveFullDataAlignment(traceMapping, cfAlignment);
			return new EmptyTraceResult() {

				public DataAlignedTrace getDataAlignment() {
					return dataAlignment;
				}

				public ControlFlowAlignmentResult getControlFlowAlignment() {
					return cfAlignment;
				}
			};
		} catch (ControlFlowAlignmentException | DataAlignmentException e) {
			String rootCause;
			if (e.getCause() != null && e.getCause().getMessage() != null) {
				rootCause = e.getCause().getMessage();
			} else {
				rootCause = e.toString();
			}
			String errorMessage = "Failed to compute the alignment between the empty trace and the model."
					+ "As possible reason is that input model is not relaxed data sound, that is, there is not a single execution trace that can reach one of the final markings from the initial marking. "
					+ "Check whether the supplied initial marking "
					+ initMarking.toString()
					+ " and final markings "
					+ Joiner.on(';').join(finalMarkings)
					+ " are valid. Another possible reason is that one of the guard expressions is invalid, that is, it does not evaluate to a Boolean value."
					+ "\n\n" + "The underlying error message was:" + rootCause;
			throw new ControlFlowAlignmentException(errorMessage, e);
		}
	}

	protected abstract AbstractPDelegate<T> getPDelegate(PetrinetGraph net, XLog log, XEventClasses classes,
			TransEvClassMapping map, Map<Transition, Integer> mapTrans2Cost, Map<XEventClass, Integer> mapEvClass2Cost,
			int delta, boolean allMarkingsAreFinal, Marking[] finalMarkings);

	/**
	 * Converts the XTrace to a LinearTrace, getting the list of event class and
	 * recording the indexes of non-mapped event classes.
	 * 
	 * @param trace
	 * @param traceMapping
	 * @param classes
	 * @param mapEvClass2Trans
	 * @param listMoveOnLog
	 * @return
	 * @throws DataAlignmentException
	 */
	protected Trace convertToLinearTrace(AbstractPDelegate<T> delegate, TraceMapping traceMapping) {
		List<XEvent> mappedTrace = traceMapping.getMappedTrace();
		int s = mappedTrace.size();
		TIntList activities = new TIntArrayList(s);
		for (XEvent event : mappedTrace) {
			int act = delegate.getIndexOf(delegate.getClassOf(event));
			boolean isMapped = act >= 0;
			assert isMapped : "Should only contain mapped classes";
			activities.add(act);

		}
		return new LinearTrace(traceMapping.getTraceName(), activities);
	}

	/**
	 * Converts the XTrace to a LinearTrace, getting the list of event class and
	 * recording the indexes of non-mapped event classes.
	 * 
	 * @param trace
	 * @param traceMapping
	 * @param classes
	 * @param mapEvClass2Trans
	 * @param listMoveOnLog
	 * @return
	 * @throws DataAlignmentException
	 */
	protected Trace convertToPartialOrderedTrace(AbstractPDelegate<T> delegate, TraceMapping traceMapping) {
		List<XEvent> mappedTrace = traceMapping.getMappedTrace();
		int s = mappedTrace.size();
		PartiallyOrderedTrace trace = new PartiallyOrderedTrace(traceMapping.getTraceName(), s);

		//		TIntList activities = new TIntArrayList(s);
		int e = 0;
		int currentTimeIndex = 0;
		long currentTime = -1;
		for (XEvent event : mappedTrace) {
			int act = delegate.getIndexOf(delegate.getClassOf(event));
			boolean isMapped = act >= 0;
			assert isMapped : "Should only contain mapped classes";
			trace.set(e, act);
			//			activities.add(act);

			long newTime = XTimeExtension.instance().extractTimestamp(event).getTime();
			if (currentTime < 0) {
				currentTime = newTime;
			}
			if (newTime > currentTime) {
				int[] pre = new int[e - currentTimeIndex];
				for (int i = 0; i < pre.length; i++) {
					pre[i] = currentTimeIndex;
					currentTimeIndex++;
				}
				trace.setPredecessors(e, pre);
				currentTime = newTime;
				assert currentTimeIndex == e;
			} else if (trace.getPredecessors(currentTimeIndex) != null) {
				trace.setPredecessors(e, trace.getPredecessors(currentTimeIndex));
			}
			e++;

		}
		return trace;
	}

	protected Map<XEvent, XEventClass> buildEventClassMap(XTrace trace) {
		Map<XEvent, XEventClass> mapEv2EventClass = new HashMap<>();
		for (XEvent event : trace) {
			mapEv2EventClass.put(event, delegate.getClassOf(event));
		}
		return mapEv2EventClass;
	}

	@SuppressWarnings("rawtypes")
	public void close() {
		if (delegate instanceof AbstractPILPDelegate) {
			((AbstractPILPDelegate) delegate).deleteLPs();
		}
	}

	protected Strategy<DRecord> getStateGraphStrategy(TraceMapping traceMapping, DataAlignmentAdapter adapter) {
		if (adapter instanceof DataAlignmentAdapterWithoutPartialImpl) { //TODO read configuration instead of instanceof
			return new DRecord.DataView.StagedStrategyImpl();
		} else {
			// calculate indices for events considering possibly events with duplicate values
			Short2ShortMap eventIndexMap = new Short2ShortOpenHashMap(traceMapping.getMappedTrace().size(),
					Hash.VERY_FAST_LOAD_FACTOR);
			eventIndexMap.defaultReturnValue((short) -1);
			Object2ShortMap<Set<XAttribute>> duplicateWritesMap = new Object2ShortOpenHashMap<>(traceMapping
					.getMappedTrace().size(), Hash.VERY_FAST_LOAD_FACTOR);
			duplicateWritesMap.defaultReturnValue((short) -1);
			int eventIndex = 0;
			int duplicateIndex = 0;
			for (XEvent e : traceMapping.getMappedTrace()) {
				Set<XAttribute> writtenAttributes = new HashSet<>();
				for (String attrName : logMapping.getReverseVariableMapping().keys()) {
					XAttribute attr = e.getAttributes().get(attrName);
					if (attr != null) {
						writtenAttributes.add(attr);
					}
				}
				short index = duplicateWritesMap.getShort(writtenAttributes);
				if (index == duplicateWritesMap.defaultReturnValue()) {
					index = (short) duplicateIndex++;
					duplicateWritesMap.put(writtenAttributes, index);
					eventIndexMap.put((short) eventIndex, index);
				} else {
					eventIndexMap.put((short) eventIndex, index);
				}
				eventIndex++;
			}
			return new DRecord.DataView.BalancedStrategyImpl(dataViewBuilder, eventIndexMap);
		}
	}

}
