package org.processmining.plugins.balancedconformance.controlflow.override;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import nl.tue.astar.Tail;
import nl.tue.astar.impl.State;
import nl.tue.astar.util.ShortShortMultiset;
import nl.tue.storage.CompressedStore;
import nl.tue.storage.StorageException;

import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.astar.petrinet.impl.PHeadCompressor;

final class DHeadCompressor<T extends Tail> extends PHeadCompressor<T> {

	public DHeadCompressor(short places, short activities) {
		super(places, activities);
	}

	@Override
	public DHead inflate(InputStream stream) throws IOException {
		// skip the hashCode
		int hashCode = readIntFromStream(stream);
		// read the marking
		byte[] mask1 = readRawMask(stream, bytesPlaces);
		ShortShortMultiset marking = inflateContentUsingRawMask(stream, mask1, places);
		// read the vector
		byte[] mask2 = readRawMask(stream, bytesActivities);
		ShortShortMultiset parikh = inflateContentUsingRawMask(stream, mask2, activities);
		assert DHead.getHashCodeProvider().hash(marking, parikh) == hashCode;
		return new DHead(marking, parikh, hashCode);
	}

	private static byte[] readRawMask(InputStream stream, int numBytes) throws IOException {
		byte[] rawMask = new byte[numBytes];
		stream.read(rawMask);
		return rawMask;
	}

	private ShortShortMultiset inflateContentUsingRawMask(InputStream stream, byte[] rawMask, int numBits)
			throws IOException {
		ShortShortMultiset result = new ShortShortMultiset((short) numBits);
		for (int i = 0; i < numBits; i++) { // ignore all remaining bits
			byte c = rawMask[i >> 3]; // div 8
			int bit = i & 7; // mod 8
			if ((c & (1 << bit)) != 0) { // bit is set
				short val = readShortFromStream(stream);
				result.put((short) i, val);
			}
		}
		return result;
	}

	@Override
	public void deflate(PHead object, OutputStream stream) throws IOException {
		// cache the hashcode for quick lookup
		writeIntToByteArray(stream, object.hashCode());
		// store the marking
		deflate(object.getMarking(), stream, places);
		// store the parikh vector
		deflate(object.getParikhVector(), stream, activities);
	}

	@Override
	public boolean equals(State<PHead, T> vector, CompressedStore<State<PHead, T>> store, long l)
			throws StorageException, IOException {
		final InputStream stream = store.getStreamForObject(l);
		final PHead head = vector.getHead();
		if (!hashEquals(head, stream)) {
			return false;
		}
		if (!multiSetEquals(head.getMarking(), stream, bytesPlaces, places)) {
			return false;
		}
		if (!multiSetEquals(head.getParikhVector(), stream, bytesActivities, activities)) {
			return false;
		}
		return true;
	}

	private boolean hashEquals(PHead head, InputStream stream) throws IOException {
		final int hashCode = readIntFromStream(stream);
		return hashCode == head.hashCode();
	}

	private boolean multiSetEquals(ShortShortMultiset multiSet, InputStream stream, int numBytes, int numBits)
			throws IOException {
		final byte[] rawMask = new byte[numBytes];
		stream.read(rawMask);
		final short[] multisetValues = multiSet.getInternalValues();
		if (!sameBitmask(multisetValues, rawMask, numBits)) {
			return false;
		}
		if (!sameValuesAssumingSameBitmask(multisetValues, rawMask, numBits, stream)) {
			return false;
		}
		return true;
	}

	private boolean sameValuesAssumingSameBitmask(short[] multisetValues, byte[] rawMask, int numBits,
			InputStream stream) throws IOException {
		for (int i = 0; i < numBits; i++) { // ignore all remaining bits
			byte c = rawMask[i >> 3]; // div 8
			int bit = i & 7; // mod 8
			short valueFromMarking = multisetValues[i];
			if ((c & (1 << bit)) != 0) {
				short valueFromStream = readShortFromStream(stream);
				if (valueFromMarking != valueFromStream) {
					return false;
				}
			}
		}
		return true;
	}

	private static boolean sameBitmask(short[] multisetValues, byte[] rawMask, int numBits) {
		for (int i = 0; i < numBits; i++) { // ignore all remaining bits
			byte c = rawMask[i >> 3]; // div 8
			int bit = i & 7; // mod 8
			if ((c & (1 << bit)) != 0) { // bit is set -> multiset should have at least one element 
				if (multisetValues[i] == 0) {
					return false;
				}
			} else { // bit is unset -> multiset should not have an element
				if (multisetValues[i] != 0) {
					return false;
				}
			}
		}
		return true;
	}

}