package org.processmining.plugins.balancedconformance.controlflow.override;

import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.shorts.Short2ShortMap;

import java.util.Arrays;

import nl.tue.astar.AStarThread;
import nl.tue.astar.Delegate;
import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.AbstractAStarThread;

import org.processmining.plugins.astar.petrinet.impl.PRecord;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataState;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;

/**
 * Similar to {@link PRecord}, but with added data costs and storing the current
 * variable assignment.
 * 
 * @author F. Mannhardt
 * 
 */
public interface DRecord extends Record {

	public final class DataView {

		private DataView() {
		}

		private static final long fmix64(long k) {
			k ^= k >>> 33;
			k *= 0xff51afd7ed558ccdL;
			k ^= k >>> 33;
			k *= 0xc4ceb9fe1a85ec53L;
			k ^= k >>> 33;
			return k;
		}

		private static int murmur3(byte[] data, int len, int seed) {
			int h1 = seed;
			int roundedEnd = (len & 0xfffffffc); // round down to 4 byte block

			h1 = mumurBody(data, h1, roundedEnd);

			// tail
			h1 = murmurTail(data, len, h1, roundedEnd);

			// finalization
			h1 ^= len;

			// fmix(h1);
			h1 = fmix(h1);

			return h1;
		}

		private static int mumurBody(byte[] data, int h1, int roundedEnd) {
			final int c1 = 0xcc9e2d51;
			final int c2 = 0x1b873593;

			for (int i = 0; i < roundedEnd; i += 4) {
				// little endian load order
				int k1 = (data[i] & 0xff) | ((data[i + 1] & 0xff) << 8) | ((data[i + 2] & 0xff) << 16)
						| (data[i + 3] << 24);
				k1 *= c1;
				k1 = (k1 << 15) | (k1 >>> 17); // ROTL32(k1,15);
				k1 *= c2;

				h1 ^= k1;
				h1 = (h1 << 13) | (h1 >>> 19); // ROTL32(h1,13);
				h1 = h1 * 5 + 0xe6546b64;
			}
			return h1;
		}

		private static int murmurTail(byte[] data, int len, int h1, int roundedEnd) {

			final int c1 = 0xcc9e2d51;
			final int c2 = 0x1b873593;

			int k1 = 0;

			switch (len & 0x03) {
				case 3 :
					k1 = (data[roundedEnd + 2] & 0xff) << 16;
					//$FALL-THROUGH$
				case 2 :
					k1 |= (data[roundedEnd + 1] & 0xff) << 8;
					//$FALL-THROUGH$
				case 1 :
					k1 |= (data[roundedEnd] & 0xff);
					k1 *= c1;
					k1 = (k1 << 15) | (k1 >>> 17); // ROTL32(k1,15);
					k1 *= c2;
					h1 ^= k1;
			}
			return h1;
		}

		private static int fmix(int h1) {
			h1 ^= h1 >>> 16;
			h1 *= 0x85ebca6b;
			h1 ^= h1 >>> 13;
			h1 *= 0xc2b2ae35;
			h1 ^= h1 >>> 16;
			return h1;
		}

		private static int addLong(byte[] buffer, long l, int pos) {
			buffer[pos++] = (byte) (l >> 56);
			buffer[pos++] = (byte) (l >> 48);
			buffer[pos++] = (byte) (l >> 40);
			buffer[pos++] = (byte) (l >> 32);
			buffer[pos++] = (byte) (l >> 24);
			buffer[pos++] = (byte) (l >> 16);
			buffer[pos++] = (byte) (l >> 8);
			buffer[pos++] = (byte) l;
			return pos;
		}

		@SuppressWarnings("unused")
		private static int addInt(byte[] buffer, int i, int pos) {
			buffer[pos++] = (byte) (i >> 24);
			buffer[pos++] = (byte) (i >> 16);
			buffer[pos++] = (byte) (i >> 8);
			buffer[pos++] = (byte) i;
			return pos;
		}

		public final static class StagedStrategyImpl implements Hash.Strategy<DRecord> {

			public int hashCode(DRecord r) {
				return (int) fmix64(r.getState());
			}

			public boolean equals(DRecord a, DRecord b) {
				if (a == b)
					return true;
				if (a == null || b == null)
					return false;
				return a.getState() == b.getState();
			}

		}

		public final static class BalancedStrategyImpl implements Hash.Strategy<DRecord> {

			private static final boolean CONSIDER_EVENT_VALUES = true;

			private static final int SEED = 89352897;

			private final DataViewBuilder dataViewBuilder;
			private final Short2ShortMap eventIndexMap;

			public BalancedStrategyImpl(DataViewBuilder dataViewBuilder, Short2ShortMap eventIndexMap) {
				this.dataViewBuilder = dataViewBuilder;
				this.eventIndexMap = eventIndexMap;
			}

			private byte[] hashBuffer = new byte[8 + 4 * 512]; // 8 byte for state, 2 bytes for 512 moves (initially)

			public int hashCode(DRecord r) {
				// else compute and cache for performance reasons
				if (!dataViewBuilder.hasWriteOps() || r.getPredecessor() == null) {
					// no data perspective or root node
					return (int) fmix64(r.getState());
				} else {
					if (r.getHashCode() != 0) {
						return r.getHashCode();
					}
					int hash = calcHash(r, r.getBacktraceSize() + 1);
					r.setHashCode(hash);
					return hash;
				}
			}

			private int calcHash(DRecord r, int estimateWriteSteps) {
				byte[] buffer = expand(hashBuffer, 8 + 4 * estimateWriteSteps);
				int pos = addLong(buffer, r.getState(), 0);
				// add sequence of moves to data state
				pos = addWriteOperations(r, buffer, pos);
				int hash = murmur3(buffer, pos, SEED);
				return hash;
			}

			private static byte[] expand(byte[] buffer, int requiredSize) {
				if (requiredSize <= buffer.length) {
					return buffer;
				}
				int newLength = Math.max(requiredSize, (int) Math.ceil(buffer.length * 2));
				if (newLength >= Integer.MAX_VALUE - 8) {
					throw new RuntimeException("Cannot allocate buffer of size" + (Integer.MAX_VALUE - 8));
				}
				return new byte[newLength];
			}

			private int addWriteOperations(DRecord r, byte[] buffer, int pos) {
				DRecord curRec = r;
				while (curRec.getPredecessor() != null) { // initial state does nothing
					int modelMove = curRec.getModelMove();
					int logMove = curRec.getMovedEvent();
					if (modelMove != AbstractAStarThread.NOMOVE) {
						short v = dataViewBuilder.buildWriteDataView(modelMove);
						if (v != -1 && logMove != AStarThread.NOMOVE) {
							// Move in both: add index of transition together with mapped event  
							v = dataViewBuilder.buildEventDataView(v, getEventIndex(logMove));
						}
						short g = dataViewBuilder.buildGuardDataView(modelMove);
						if (v != -1 || g != -1) {
							buffer[pos++] = (byte) (v >> 8);
							buffer[pos++] = (byte) v;
							buffer[pos++] = (byte) (g >> 8);
							buffer[pos++] = (byte) g;
						}
					}
					curRec = curRec.getPredecessor();
				}
				return pos;
			}

			public boolean equals(DRecord a, DRecord b) {
				if (a == b)
					return true;
				if (a == null || b == null)
					return false;
				if (a.getState() != b.getState())
					return false;
				if (a.getPredecessor() == null && b.getPredecessor() == null) { // initial states records nothing
					return true;
				}

				DRecord curA = a, curB = b;

				while (curA.getPredecessor() != null || curB.getPredecessor() != null) {

					short vA = 0;
					short vB = 0;
					short gA = 0;
					short gB = 0;

					while (curA.getPredecessor() != null) {
						int modelMove = curA.getModelMove();
						int logMove = curA.getMovedEvent();
						curA = curA.getPredecessor();
						if (modelMove != AStarThread.NOMOVE) {
							short v = dataViewBuilder.buildWriteDataView(modelMove);
							if (v != -1) {
								if (logMove != AStarThread.NOMOVE) {
									// Move in both: add index of transition together with mapped event  
									v = dataViewBuilder.buildEventDataView(v, getEventIndex(logMove));
								} // else free variable
							}
							short g = dataViewBuilder.buildGuardDataView(modelMove);
							if (v != -1 || g != -1) {
								vA = v;
								gA = g;
								break;
							} // else ignore since not part of data perspective 
						}
					}

					while (curB.getPredecessor() != null) {
						int modelMove = curB.getModelMove();
						int logMove = curB.getMovedEvent();
						curB = curB.getPredecessor();
						if (modelMove != AStarThread.NOMOVE) {
							short v = dataViewBuilder.buildWriteDataView(modelMove);
							if (v != -1) {
								if (logMove != AStarThread.NOMOVE) {
									// Move in both: add index of transition together with mapped event  
									v = dataViewBuilder.buildEventDataView(v, getEventIndex(logMove));
								} // else free variable
							}
							short g = dataViewBuilder.buildGuardDataView(modelMove);
							if (v != -1 || g != -1) {
								vB = v;
								gB = g;
								break;
							} // else ignore since not part of data perspective
						}
					}

					if (vA != vB || gA != gB) {
						assert !Arrays.equals(getDataView(dataViewBuilder, a), getDataView(dataViewBuilder, b)) : "Invalid equals, data view should be different:\nA: "
								+ Arrays.toString(getDataView(dataViewBuilder, a))
								+ "\nB: "
								+ Arrays.toString(getDataView(dataViewBuilder, b));
						return false;
					}
				}
				assert Arrays.equals(getDataView(dataViewBuilder, a), getDataView(dataViewBuilder, b)) : "Invalid equals, data view should be same:\nA: "
						+ Arrays.toString(getDataView(dataViewBuilder, a))
						+ "\nB: "
						+ Arrays.toString(getDataView(dataViewBuilder, b));
				return true;
			}

			private short getEventIndex(int logMove) {
				if (CONSIDER_EVENT_VALUES) {
					short result = eventIndexMap.get((short) logMove);
					if (result != eventIndexMap.defaultReturnValue()) {
						return result;
					} else {
						throw new RuntimeException("Event " + logMove + " did not write anything!");
					}
				} else {
					return (short) logMove;
				}
			}

			private static final short[] EMPTY = new short[0];

			private short[] getDataView(DataViewBuilder builder, DRecord record) {
				if (record.getPredecessor() == null) {
					// nothing can have been written
					return EMPTY;
				}
				short[] dataView = new short[(record.getBacktraceSize() + 1) * 2];
				DRecord curRec = record;
				int i = 0;
				// add sequence of moves to data state
				while (curRec.getPredecessor() != null) {
					int modelMove = curRec.getModelMove();
					int logMove = curRec.getMovedEvent();
					if (modelMove != AStarThread.NOMOVE) {
						short tIndex = builder.buildWriteDataView(modelMove);
						if (tIndex != -1) {
							if (logMove != AStarThread.NOMOVE) {
								short eIndex = builder.buildEventDataView(tIndex, getEventIndex(logMove));
								if (eIndex != -1) {
									dataView[i++] = eIndex;
								} else {
									dataView[i++] = tIndex;
								}
							} else {
								dataView[i++] = tIndex;
							}
						}
						short gIndex = builder.buildGuardDataView(modelMove);
						if (gIndex != -1) {
							dataView[i++] = gIndex;
						}
					}
					curRec = curRec.getPredecessor();
				}
				return Arrays.copyOf(dataView, i);
			}

		}

	}

	@Override
	public DRecord getNextRecord(Delegate<? extends Head, ? extends Tail> d, Trace trace, Head newHead, long state,
			int modelMove, int movedEvent, int activity);

	@Override
	public DRecord getPredecessor();

	public int getNextEvent(Delegate<? extends Head, ? extends Tail> delegate, Trace trace);

	int getExecutedEvents();

	public int getTotalIntegerCost();

	public int getCost();

	public int getDataCost();

	public void setDataCost(int dataCost);

	public int getDataState();

	public void setDataState(int dataState);

	DataState getDataState(DataStateStore store);

	public DRecord[] getHistory();

	public void setHashCode(int hashCode);

	public int getHashCode();

}