/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.controlflow.resultbuilder;

import java.util.Map;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XEvent;
import org.processmining.log.utils.XUtils;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.astar.petrinet.impl.PHead;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResultImpl;
import org.processmining.plugins.balancedconformance.controlflow.UnreliableControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.controlflow.UnreliableControlFlowAlignmentException.UnreliableReason;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;

import nl.tue.astar.AStarException;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.AStarThread.Canceller;
import nl.tue.astar.Tail;
import nl.tue.astar.Trace;

public final class AlignmentResultBuilderImpl implements AlignmentResultBuilder {

	private final DataAwareThread<PHead, ? extends Tail> thread;
	private final Trace linearTrace;
	private final TraceMapping traceMapping;

	private final Canceller canceller;
	private final AbstractPDelegate<? extends Tail> delegate;
	private final Map<XEventClass, Integer> mapEvClass2Cost;
	private final Map<XEvent, XEventClass> mapEv2EventClass;

	boolean hasReturnedOptimal = false;
	private double timeLimit;
	private double delta;
	private AStarObserver observer;	

	public AlignmentResultBuilderImpl(DataAwareThread<PHead, ? extends Tail> thread, Trace linearTrace,
			TraceMapping traceMapping, Canceller canceller, AbstractPDelegate<? extends Tail> delegate,
			Map<XEvent, XEventClass> mapEv2EventClass, Map<XEventClass, Integer> mapEvClass2Cost, double timeLimit,
			AStarObserver observer) {
		this.thread = thread;
		this.linearTrace = linearTrace;
		this.traceMapping = traceMapping;
		this.canceller = canceller;
		this.delegate = delegate;
		this.timeLimit = timeLimit;
		this.observer = observer;
		this.delta = delegate.getDelta();
		this.mapEv2EventClass = mapEv2EventClass;
		this.mapEvClass2Cost = mapEvClass2Cost;
		if (observer != null) {
			thread.addObserver(observer);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.DataConformance.DataAlignment.PetriNet.Plus
	 * .ControlFlow.ControlFlowAlignmentResultBuilder#nextResult()
	 */
	public ControlFlowAlignmentResult calcResult(double upperBound)
			throws ControlFlowAlignmentException, UnreliableControlFlowAlignmentException {
		try {

			int scaledUpperBound = (int) Math.floor((upperBound * delta) + (delta - 1));
			if (upperBound == Double.MAX_VALUE) {
				scaledUpperBound = Integer.MAX_VALUE;
			}
			DRecord record = (DRecord) thread.getOptimalRecord(canceller, scaledUpperBound, getTimeLimit());
			if (thread.wasReliable()) {
				hasReturnedOptimal = true;
				if (observer != null) {
					observer.close();
				}
				return createAlignmentResult(record);
			} else if (canceller.isCancelled()) {
				throw new UnreliableControlFlowAlignmentException("Cancelled",
						record != null ? createAlignmentResult(record) : null, UnreliableReason.CANCELLED);
			} else {
				if (thread.isQueueEmpty()) {
					//TODO it might still be maxCost related
					throw new UnreliableControlFlowAlignmentException(
							String.format(
									"Could not compute an alignment for trace %3$s after visiting %2$s states and queueing %1$s states.",
									thread.getQueuedStateCount(), thread.getVisitedStateCount(),
									XUtils.getConceptName(traceMapping.getTrace())),
							record != null ? createAlignmentResult(record) : null, UnreliableReason.IMPOSSIBLE);

				} else {
					throw new UnreliableControlFlowAlignmentException(
							String.format(
									"Maximum alignment cost or maximum number of queued states reached for trace %3$s after visiting %2$s states and queueing %1$s states.",
									thread.getQueuedStateCount(), thread.getVisitedStateCount(),
									XUtils.getConceptName(traceMapping.getTrace())),
							record != null ? createAlignmentResult(record) : null, UnreliableReason.MAXCOST);
				}
			}
		} catch (AStarException e) {
			if (observer != null) {
				observer.close();
			}
			throw new ControlFlowAlignmentException(e);
		}
	}

	private ControlFlowAlignmentResultImpl createAlignmentResult(DRecord record) {
		return new ControlFlowAlignmentResultImpl(record, delegate, mapEv2EventClass, mapEvClass2Cost,
				traceMapping.getTrace(), traceMapping.getTrace(), linearTrace);
	}

	public long getTotalQueuedStates() {
		return thread.getQueuedStateCount();
	}

	public int getPartialDataAlignments() {
		return thread.getPartialDataAlignments();
	}

	public int getCacheSize() {
		return thread.getCacheSize();
	}

	public int getCacheHit() {
		return thread.getCacheHit();
	}

	public boolean hasNext() {
		return !hasReturnedOptimal;
	}

	public long getDataStateCount() {
		return thread.getDataStateCount();
	}

	public double getTimeLimit() {
		return timeLimit;
	}

	public void setTimeLimit(double timeLimit) {
		this.timeLimit = timeLimit;
	}

}