package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Map;

import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateImpl.VariableIndexer;
import org.processmining.plugins.balancedconformance.dataflow.datastate.compression.Compressor;
import org.processmining.plugins.balancedconformance.dataflow.datastate.compression.DeltaCompressor;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.Ordering;

/**
 * Provider for Map based DataStates
 */
public final class DataStateFactoryImpl implements DataStateFactory {

	final static class VariableIndexerImpl implements VariableIndexer {

		@SuppressWarnings("rawtypes")
		private static Ordering<Class> CLASS_ORDER = Ordering.<Class>explicit(Boolean.class, String.class,
				Integer.class, Long.class, Date.class, Float.class, Double.class);

		private BiMap<String, Integer> indexMap;

		private final int sizeWithPrimes;
		private final int normalSize;
		private final int sizeForBitset;
		private final int overallArraySize;

		public VariableIndexerImpl(@SuppressWarnings("rawtypes") final Map<String, Class> variableTypes) {
			this.indexMap = HashBiMap.create(variableTypes.size());
			int index = 0;
			List<String> varSet = new ArrayList<>(variableTypes.keySet());
			Collections.sort(varSet, Ordering.from(new Comparator<String>() {

				@SuppressWarnings("rawtypes")
				public int compare(String o1, String o2) {
					Class class1 = variableTypes.get(o1);
					Class class2 = variableTypes.get(o2);
					return CLASS_ORDER.compare(class1, class2);
				}
			}).compound(Ordering.natural()));
			for (String varName : varSet) {
				indexMap.put(varName.intern(), index++);
			}
			this.normalSize = indexMap.size();
			for (String varName : varSet) {
				indexMap.put(varName.concat("'").intern(), index++);
			}
			this.sizeWithPrimes = indexMap.size();
			this.sizeForBitset = ((sizeWithPrimes - 1) >> 6) + 1;
			this.overallArraySize = sizeWithPrimes + sizeForBitset;
		}

		public int sizeWithPrimes() {
			return sizeWithPrimes;
		}

		public int sizeNormal() {
			return normalSize;
		}

		public int getNormalIndex(int primeIndex) {
			return (primeIndex) - normalSize;
		}

		public int getPrimeIndex(int normalIndex) {
			return normalIndex + normalSize;
		}

		public int getIndex(String varName) {
			return indexMap.get(varName);
		}

		public int getFirstPrimeIndex() {
			return normalSize;
		}

		public String getVarName(int index) {
			return indexMap.inverse().get(index);
		}

		public int sizeIncludingBitset() {
			return overallArraySize;
		}

		public int sizeBitset() {
			return sizeForBitset;
		}

	}

	private final VariableIndexer indexer;
	private final Compressor compressor;

	public DataStateFactoryImpl(@SuppressWarnings("rawtypes") Map<String, Class> variableTypes) {
		this(variableTypes, new DeltaCompressor());
	}

	public DataStateFactoryImpl(@SuppressWarnings("rawtypes") Map<String, Class> variableTypes, Compressor compressor) {
		this.compressor = compressor;
		indexer = new VariableIndexerImpl(variableTypes);
	}

	public DataState newDataState(int initialCapacity) {
		return new DataStateImpl(indexer);
	}

	public DataState newDataState() {
		return new DataStateImpl(indexer);
	}

	public DataState decompressFromStore(Object data) {
		long[] longData = compressor.decompress((byte[]) data, indexer.sizeIncludingBitset());
		return new DataStateImpl(indexer, longData);
	}

	public Object compressToStore(DataState state) {
		return compressor.compress((long[]) state.getData());
	}

}