/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.LockSupport;

import org.cliffc.high_scale_lib.NonBlockingHashMap;
import org.cliffc.high_scale_lib.NonBlockingHashMapLong;

/**
 * Implementation of an {@link DataStateStore} using the
 * {@link NonBlockingHashMap} together with a compare-and-swap operation.
 * 
 * @author F. Mannhardt
 * 
 */
public final class DataStateStoreConcurrentImpl implements DataStateStore {

	private final NonBlockingHashMap<DataState, Integer> dataStateStore;
	private final NonBlockingHashMapLong<DataState> reverseDataStateStore;
	private final AtomicInteger keyCounter = new AtomicInteger(START_INDEX);

	public DataStateStoreConcurrentImpl(int initialCapacity) {
		super();
		dataStateStore = new NonBlockingHashMap<>(initialCapacity);
		reverseDataStateStore = new NonBlockingHashMapLong<>(initialCapacity, false);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #get(long)
	 */
	public DataState get(int index) {
		if (index == EMPTY_STATE_INDEX) {
			return EMPTY_IMMUTABLE_STATE;
		}
		return reverseDataStateStore.get(index);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #putIfAbsent
	 * (org.processmining.plugins.balancedconformance.dataflow.DataState)
	 */
	public int putIfAbsent(DataState state) {
		if (state == EMPTY_IMMUTABLE_STATE) {
			return EMPTY_STATE_INDEX;
		}
		state = state.makeStateImmutable();
		Integer idx = dataStateStore.putIfAbsent(state, NO_ENTRY);
		if (idx == null) {
			idx = keyCounter.getAndIncrement();
			reverseDataStateStore.put(idx.longValue(), state);
			dataStateStore.put(state, idx);
			return idx;
		} else {
			while (idx == NO_ENTRY) {
				idx = dataStateStore.get(state);
				LockSupport.parkNanos(10L);
			}
			return idx;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #size()
	 */
	public int size() {
		return dataStateStore.size();
	}

}
