/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.Arrays;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import it.unimi.dsi.fastutil.Hash.Strategy;
import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenCustomHashMap;

/**
 * {@link DataStateStore} implementation using {@link Object2IntMap} maps &
 * {@link ReentrantReadWriteLock} and compressed {@link DataState}s.
 * 
 * @author F. Mannhardt
 * 
 */
public final class DataStateStoreMemoryOptimizedImpl implements DataStateStore {

	private int stateCounter = START_INDEX;

	private final Object2IntMap<Object> dataStateStore;
	private final Int2ObjectMap<Object> reverseDataStateStore;
	private final ReentrantReadWriteLock lock = new ReentrantReadWriteLock();
	private final Lock readLock = lock.readLock();
	private final Lock writeLock = lock.writeLock();

	private final DataStateFactory factory;

	public DataStateStoreMemoryOptimizedImpl(DataStateFactory factory, int initialCapacity) {
		super();
		this.factory = factory;
		dataStateStore = new Object2IntOpenCustomHashMap<>(initialCapacity, 0.5f, new Strategy<Object>() {

			public int hashCode(Object o) {
				return Arrays.hashCode((byte[]) o);
			}

			public boolean equals(Object a, Object b) {
				return Arrays.equals((byte[]) a, (byte[]) b);

			}
		});
		dataStateStore.defaultReturnValue(NO_ENTRY);
		reverseDataStateStore = new Int2ObjectOpenHashMap<>(initialCapacity, 0.5f);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #get(long)
	 */
	public DataState get(int index) {
		if (index == EMPTY_STATE_INDEX) {
			return EMPTY_IMMUTABLE_STATE;
		}
		Object data = null;
		readLock.lock();
		try {
			data = reverseDataStateStore.get(index);
		} finally {
			readLock.unlock();
		}
		if (data != null) {
			return factory.decompressFromStore(data);
		} else {
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #putIfAbsent
	 * (org.processmining.plugins.balancedconformance.dataflow.DataState)
	 */
	public int putIfAbsent(DataState state) {
		if (state == EMPTY_IMMUTABLE_STATE) {
			return EMPTY_STATE_INDEX;
		}
		Object data = factory.compressToStore(state);
		readLock.lock();
		try {
			int idx = dataStateStore.getInt(data);
			if (idx != NO_ENTRY) {
				return idx;
			}
		} finally {
			readLock.unlock();
		}
		writeLock.lock();
		try {
			int idx = dataStateStore.getInt(data);
			if (idx != NO_ENTRY) {
				return idx;
			} else {
				int stateId = stateCounter++;
				dataStateStore.put(data, stateId);
				reverseDataStateStore.put(stateId, data);
				return stateId;
			}
		} finally {
			writeLock.unlock();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #size()
	 */
	public int size() {
		readLock.lock();
		try {
			return dataStateStore.size();
		} finally {
			readLock.unlock();
		}
	}

}
