/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.Arrays;

import it.unimi.dsi.fastutil.Hash.Strategy;
import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenCustomHashMap;

/**
 * {@link DataStateStore} implementation using thread-UNSAFE FastUtil maps and
 * compressed {@link DataState}s.
 * 
 * @author F. Mannhardt
 * 
 */
public final class DataStateStoreMemoryOptimizedNoLockImpl implements DataStateStore {

	private int stateCounter = START_INDEX;

	private final Object2IntMap<Object> dataStateStore;
	private final Int2ObjectMap<Object> reverseDataStateStore;

	private final DataStateFactory factory;

	public DataStateStoreMemoryOptimizedNoLockImpl(DataStateFactory factory, int initialCapacity) {
		super();
		this.factory = factory;
		dataStateStore = new Object2IntOpenCustomHashMap<>(initialCapacity, 0.5f, new Strategy<Object>() {

			public int hashCode(Object o) {
				return Arrays.hashCode((byte[]) o);
			}

			public boolean equals(Object a, Object b) {
				return Arrays.equals((byte[]) a, (byte[]) b);

			}
		});
		dataStateStore.defaultReturnValue(NO_ENTRY);
		reverseDataStateStore = new Int2ObjectOpenHashMap<>(initialCapacity, 0.5f);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #get(long)
	 */
	public DataState get(int index) {
		if (index == EMPTY_STATE_INDEX) {
			return EMPTY_IMMUTABLE_STATE;
		}
		Object data = reverseDataStateStore.get(index);
		if (data != null) {
			return factory.decompressFromStore(data);
		} else {
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #putIfAbsent
	 * (org.processmining.plugins.balancedconformance.dataflow.DataState)
	 */
	public int putIfAbsent(DataState state) {
		if (state == EMPTY_IMMUTABLE_STATE) {
			return EMPTY_STATE_INDEX;
		}
		Object data = factory.compressToStore(state);
		int idx = dataStateStore.getInt(data);
		if (idx != NO_ENTRY) {
			return idx;
		} else {
			int stateId = stateCounter++;
			dataStateStore.put(data, stateId);
			reverseDataStateStore.put(stateId, data);
			return stateId;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.balancedconformance.dataflow.DataStateStore
	 * #size()
	 */
	public int size() {
		return dataStateStore.size();
	}

}
