package org.processmining.plugins.balancedconformance.functions;

import java.util.Map.Entry;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.datapetrinets.expression.syntax.ExprAnd;
import org.processmining.datapetrinets.expression.syntax.ExprAtLeast;
import org.processmining.datapetrinets.expression.syntax.ExprAtMost;
import org.processmining.datapetrinets.expression.syntax.ExprDiv;
import org.processmining.datapetrinets.expression.syntax.ExprEqual;
import org.processmining.datapetrinets.expression.syntax.ExprFunction;
import org.processmining.datapetrinets.expression.syntax.ExprGreaterThan;
import org.processmining.datapetrinets.expression.syntax.ExprLessThan;
import org.processmining.datapetrinets.expression.syntax.ExprLitBoolean;
import org.processmining.datapetrinets.expression.syntax.ExprLitDouble;
import org.processmining.datapetrinets.expression.syntax.ExprLitInteger;
import org.processmining.datapetrinets.expression.syntax.ExprLitNull;
import org.processmining.datapetrinets.expression.syntax.ExprLitString;
import org.processmining.datapetrinets.expression.syntax.ExprMinus;
import org.processmining.datapetrinets.expression.syntax.ExprMult;
import org.processmining.datapetrinets.expression.syntax.ExprNegation;
import org.processmining.datapetrinets.expression.syntax.ExprNot;
import org.processmining.datapetrinets.expression.syntax.ExprNotEqual;
import org.processmining.datapetrinets.expression.syntax.ExprOr;
import org.processmining.datapetrinets.expression.syntax.ExprPlus;
import org.processmining.datapetrinets.expression.syntax.ExprRoot;
import org.processmining.datapetrinets.expression.syntax.ExprVariable;
import org.processmining.datapetrinets.expression.syntax.ExpressionParserDefaultVisitor;
import org.processmining.datapetrinets.expression.syntax.ExpressionParserTreeConstants;
import org.processmining.datapetrinets.expression.syntax.ExpressionVisitorException;
import org.processmining.datapetrinets.expression.syntax.Node;
import org.processmining.datapetrinets.expression.syntax.SimpleNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.plugins.DataConformance.framework.WriteOperationCost;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.functions.impl.FunctionRegistry;
import org.processmining.plugins.balancedconformance.mapping.Variable;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.SetMultimap;

public final class FunctionsTransformations {

	private final static class ExpressionParserFunctionVisitor extends ExpressionParserDefaultVisitor {

		private boolean hasFunction;

		public Object visit(ExprFunction node, Object data) throws ExpressionVisitorException {
			Object value = node.jjtGetValue();
			if (value != null) {
				Function factory = FunctionRegistry.getFunction(value);
				if (factory.isValidFunction(node)) {
					hasFunction = true;
				}
			}
			return data;
		}

		public boolean isHasTimeFunction() {
			return hasFunction;
		}

	}

	private final static class ExpressionParserFunctionTransformationVisitor extends ExpressionParserDefaultVisitor {

		private final PNWDTransition transition;
		private final SetMultimap<Transition, VirtualVariable> virtualVariables;
		private final SetMultimap<String, Transition> nameToTransition;
		private final ImmutableMap<String, Variable> normalVariables;
		private final WriteOperationCost writeOperationCost;

		public ExpressionParserFunctionTransformationVisitor(PNWDTransition transition,
				SetMultimap<String, Transition> nameToTransition,
				SetMultimap<Transition, VirtualVariable> artificialVariables, ImmutableMap<String, Variable> normalVariables, WriteOperationCost writeOperationCost) {
			this.transition = transition;
			this.virtualVariables = artificialVariables;
			this.nameToTransition = nameToTransition;
			this.normalVariables = normalVariables;
			this.writeOperationCost = writeOperationCost;
		}

		public Object visit(ExprRoot node, Object data) throws ExpressionVisitorException {
			ExprRoot root = new ExprRoot(ExpressionParserTreeConstants.JJTROOT);
			root.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			return root;
		}

		public Object visit(ExprOr node, Object data) throws ExpressionVisitorException {
			ExprOr newNode = new ExprOr(ExpressionParserTreeConstants.JJTOR);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprAnd node, Object data) throws ExpressionVisitorException {
			ExprAnd newNode = new ExprAnd(ExpressionParserTreeConstants.JJTAND);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprEqual node, Object data) throws ExpressionVisitorException {
			ExprEqual newNode = new ExprEqual(ExpressionParserTreeConstants.JJTEQUAL);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprNotEqual node, Object data) throws ExpressionVisitorException {
			ExprNotEqual newNode = new ExprNotEqual(ExpressionParserTreeConstants.JJTNOTEQUAL);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprLessThan node, Object data) throws ExpressionVisitorException {
			ExprLessThan newNode = new ExprLessThan(ExpressionParserTreeConstants.JJTLESSTHAN);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprAtMost node, Object data) throws ExpressionVisitorException {
			ExprAtMost newNode = new ExprAtMost(ExpressionParserTreeConstants.JJTATMOST);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprGreaterThan node, Object data) throws ExpressionVisitorException {
			ExprGreaterThan newNode = new ExprGreaterThan(ExpressionParserTreeConstants.JJTGREATERTHAN);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprAtLeast node, Object data) throws ExpressionVisitorException {
			ExprAtLeast newNode = new ExprAtLeast(ExpressionParserTreeConstants.JJTATLEAST);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprPlus node, Object data) throws ExpressionVisitorException {
			ExprPlus newNode = new ExprPlus(ExpressionParserTreeConstants.JJTPLUS);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprMinus node, Object data) throws ExpressionVisitorException {
			ExprMinus newNode = new ExprMinus(ExpressionParserTreeConstants.JJTMINUS);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprMult node, Object data) throws ExpressionVisitorException {
			ExprMult newNode = new ExprMult(ExpressionParserTreeConstants.JJTMULT);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprDiv node, Object data) throws ExpressionVisitorException {
			ExprDiv newNode = new ExprDiv(ExpressionParserTreeConstants.JJTDIV);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			newNode.jjtAddChild((Node) node.jjtGetChild(1).jjtAccept(this, data), 1);
			return newNode;
		}

		public Object visit(ExprNegation node, Object data) throws ExpressionVisitorException {
			ExprNegation newNode = new ExprNegation(ExpressionParserTreeConstants.JJTNEGATION);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprNot node, Object data) throws ExpressionVisitorException {
			ExprNot newNode = new ExprNot(ExpressionParserTreeConstants.JJTNOT);
			newNode.jjtAddChild((Node) node.jjtGetChild(0).jjtAccept(this, data), 0);
			return newNode;
		}

		public Object visit(ExprVariable node, Object data) throws ExpressionVisitorException {
			ExprVariable newNode = new ExprVariable(ExpressionParserTreeConstants.JJTVARIABLE);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitInteger node, Object data) throws ExpressionVisitorException {
			ExprLitInteger newNode = new ExprLitInteger(ExpressionParserTreeConstants.JJTLITINTEGER);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitDouble node, Object data) throws ExpressionVisitorException {
			ExprLitDouble newNode = new ExprLitDouble(ExpressionParserTreeConstants.JJTLITDOUBLE);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitString node, Object data) throws ExpressionVisitorException {
			ExprLitString newNode = new ExprLitString(ExpressionParserTreeConstants.JJTLITSTRING);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitBoolean node, Object data) throws ExpressionVisitorException {
			ExprLitBoolean newNode = new ExprLitBoolean(ExpressionParserTreeConstants.JJTLITBOOLEAN);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprLitNull node, Object data) throws ExpressionVisitorException {
			ExprLitNull newNode = new ExprLitNull(ExpressionParserTreeConstants.JJTLITNULL);
			newNode.jjtSetValue(node.jjtGetValue());
			return newNode;
		}

		public Object visit(ExprFunction node, Object data) throws ExpressionVisitorException {

			try {
				if (node.jjtGetValue() != null) {
					Function function = FunctionRegistry.getFunction(node.jjtGetValue());
					if (function != null) {
						if (function instanceof FunctionUnary) {

							FunctionUnary unaryFunction = (FunctionUnary) function;

							String parameter1 = tryGetParameterValue(node, 0);

							VirtualVariable primaryVariable = unaryFunction.createPrimaryVariable(normalVariables, writeOperationCost, nameToTransition,
									transition, parameter1);							
							SetMultimap<Transition, VirtualVariable> helperVariables = unaryFunction
									.createHelperVariables(normalVariables, writeOperationCost, nameToTransition, transition, primaryVariable, parameter1);
							for (final Entry<Transition, VirtualVariable> entry : helperVariables.entries()) {
								// Evaluated in insertion order
								virtualVariables.put(entry.getKey(), entry.getValue());
							}
							// First helper, then primary
							virtualVariables.put(transition, primaryVariable);
							ExprVariable substituteVariable = new ExprVariable(
									ExpressionParserTreeConstants.JJTVARIABLE);
							substituteVariable.jjtSetValue(DataStateHelper.addPrime(primaryVariable.getName()));
							return substituteVariable;

						} else if (function instanceof FunctionBinary) {
							FunctionBinary binaryFunction = (FunctionBinary) function;

							String parameter1 = tryGetParameterValue(node, 0);
							String parameter2 = tryGetParameterValue(node, 1);

							VirtualVariable primaryVariable = binaryFunction.createPrimaryVariable(normalVariables, writeOperationCost, nameToTransition,
									transition, parameter1, parameter2);							
							SetMultimap<Transition, VirtualVariable> helperVariables = binaryFunction
									.createHelperVariables(normalVariables, writeOperationCost, nameToTransition, transition, primaryVariable, parameter1,
											parameter2);
							for (final Entry<Transition, VirtualVariable> entry : helperVariables.entries()) {
								// Evaluated in insertion order
								virtualVariables.put(entry.getKey(), entry.getValue());
							}
							virtualVariables.put(transition, primaryVariable);
							// First helper, then primary
							ExprVariable substituteVariable = new ExprVariable(
									ExpressionParserTreeConstants.JJTVARIABLE);
							substituteVariable.jjtSetValue(DataStateHelper.addPrime(primaryVariable.getName()));
							return substituteVariable;

						} else if (function instanceof FunctionTernary) {
							FunctionTernary ternaryFunction = (FunctionTernary) function;

							String parameter1 = tryGetParameterValue(node, 0);
							String parameter2 = tryGetParameterValue(node, 1);
							String parameter3 = tryGetParameterValue(node, 2);

							VirtualVariable primaryVariable = ternaryFunction.createPrimaryVariable(normalVariables, writeOperationCost, nameToTransition,
									transition, parameter1, parameter2, parameter3);							
							SetMultimap<Transition, VirtualVariable> helperVariables = ternaryFunction
									.createHelperVariables(normalVariables, writeOperationCost, nameToTransition, transition, primaryVariable, parameter1,
											parameter2, parameter3);
							for (final Entry<Transition, VirtualVariable> entry : helperVariables.entries()) {
								// Evaluated in insertion order
								virtualVariables.put(entry.getKey(), entry.getValue());
							}
							// First helper, then primary
							virtualVariables.put(transition, primaryVariable);
							ExprVariable substituteVariable = new ExprVariable(
									ExpressionParserTreeConstants.JJTVARIABLE);
							substituteVariable.jjtSetValue(DataStateHelper.addPrime(primaryVariable.getName()));
							return substituteVariable;
						}
					}
				}
			} catch (FunctionException e) {
				throw new ExpressionVisitorException("Invalid function " + e.getMessage(), e);
			}
			throw new ExpressionVisitorException("Invalid function " + node);

		}

		private static String tryGetParameterValue(Node node, int index) throws ExpressionVisitorException {
			Node child = tryGetChild(node, index);
			if (child != null) {
				if (child instanceof ExprLitString) {
					String unescaped = (String) ((SimpleNode) child).jjtGetValue();
					return unescaped.substring(1, unescaped.length() - 1);
				} else if (child instanceof ExprVariable) {
					return (String) ((ExprVariable) child).jjtGetValue();
				}
				throw new ExpressionVisitorException("Invalid parameter " + node);
			} else {
				return null;
			}

		}

		private static Node tryGetChild(Node node, int index) {
			if (node.jjtGetNumChildren() - 1 >= index) {
				return node.jjtGetChild(index);
			} else {
				return null;
			}
		}
	}

	private FunctionsTransformations() {
		super();
	}

	public static GuardExpression transformExpression(PNWDTransition transition,
			SetMultimap<String, Transition> nameToTransition,
			SetMultimap<Transition, VirtualVariable> virtualVariables, ImmutableMap<String, Variable> normalVariables,
			WriteOperationCost writeOperationCost) {
		try {
			ExpressionParserFunctionTransformationVisitor visitor = new ExpressionParserFunctionTransformationVisitor(
					transition, nameToTransition, virtualVariables, normalVariables, writeOperationCost);
			ExprRoot exprRoot = (ExprRoot) transition.getGuardExpression().visit(visitor, null);
			return GuardExpression.Factory.newInstance(exprRoot);
		} catch (ExpressionVisitorException e) {
			throw new RuntimeException(e);
		}
	}

	public static boolean hasFunction(PNWDTransition pnwdTransition) {
		GuardExpression expression = pnwdTransition.getGuardExpression();
		if (expression != null) {
			try {
				ExpressionParserFunctionVisitor visitor = new ExpressionParserFunctionVisitor();
				expression.visit(visitor, null);
				return visitor.isHasTimeFunction();
			} catch (ExpressionVisitorException e) {
				throw new RuntimeException(e);
			}
		}
		return false;
	}

}