package org.processmining.plugins.balancedconformance.functions.impl;

import java.util.Set;

import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.DataConformance.framework.WriteOperationCost;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateHelper;
import org.processmining.plugins.balancedconformance.functions.VariableNameGenerator;
import org.processmining.plugins.balancedconformance.functions.VirtualVariable;
import org.processmining.plugins.balancedconformance.mapping.Variable;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

public abstract class StringFunctionVariable implements VirtualVariable {

	protected final String variableName;
	protected final String value;

	protected final Transition transition;
	protected final boolean isPrime;

	private final String name;
	private final Class<?> type = Boolean.class;

	private final Variable targetVariable;
	private final boolean isFinal;
	private final int missingValueCost;
	private final int faultyValueCost;

	public StringFunctionVariable(ImmutableMap<String, Variable> normalVariables,
			WriteOperationCost writeOperationCost, Transition transition, String variableNameMaybePrime,
			String parameterValue, String optionalIdentifier) {
		this.transition = transition;
		this.isPrime = DataStateHelper.isPrimeVar(variableNameMaybePrime);
		this.variableName = DataStateHelper.stripPrimeIfNeeded(variableNameMaybePrime);
		this.value = parameterValue;
		if (optionalIdentifier != null) {
			this.name = determineName(variableNameMaybePrime, optionalIdentifier);
		} else {
			this.name = determineName(variableNameMaybePrime, VariableNameGenerator.next());
		}
		this.targetVariable = normalVariables.get(this.variableName);
		if (targetVariable != null) {
			isFinal = writeOperationCost.isFinal(transition.getLabel(), targetVariable.getName());
			missingValueCost = (int) writeOperationCost.notWritingCost(transition.getLabel(), targetVariable.getName());
			faultyValueCost = (int) writeOperationCost.costFaultyValue(transition.getLabel(), targetVariable.getName());
		} else {
			// defaults without target variable
			isFinal = true;
			missingValueCost = Integer.MAX_VALUE;
			faultyValueCost = Integer.MAX_VALUE;
		}
	}

	abstract protected String determineName(String variableName, String next);

	public String getName() {
		return name;
	}

	public Class<?> getType() {
		return type;
	}

	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((getName() == null) ? 0 : getName().hashCode());
		return result;
	}

	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof StringFunctionVariable))
			return false;
		StringFunctionVariable other = (StringFunctionVariable) obj;
		if (getName() == null) {
			if (other.getName() != null)
				return false;
		} else if (!getName().equals(other.getName()))
			return false;
		return true;
	}

	public Set<String> getRelevantAttributes() {
		return ImmutableSet.of(getTargetAttributeName());
	}

	public int getFaultyValueCost() {
		return faultyValueCost;
	}

	public int getMissingValueCost() {
		return missingValueCost;
	}

	public Double getUpperBound() {
		return 1.0d; // boolean variable 
	}

	public Set<Transition> getRelevantTransitions() {
		return ImmutableSet.of(transition);
	}

	public boolean isFinal() {
		return isFinal;
	}
	
	public boolean needsAlignedPrefix() {
		return !isPrime;
	}

	protected String getTargetAttributeName() {
		if (targetVariable != null) {
			return targetVariable.getAttributeName();
		} else {
			return variableName; // Just take the name itself
		}
	}

}
