/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.ui;

import java.awt.Color;
import java.awt.Dimension;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;

import nl.tue.astar.AStarThread.ASynchronousMoveSorting;
import nl.tue.astar.AStarThread.QueueingModel;

import org.processmining.framework.util.ui.widgets.ProMComboBox;
import org.processmining.framework.util.ui.widgets.ProMPropertiesExpertModePanel;
import org.processmining.framework.util.ui.widgets.WidgetColors;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.controlflow.adapter.SearchMethod;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter.ILPSolver;

import com.fluxicon.slickerbox.components.NiceDoubleSlider;
import com.fluxicon.slickerbox.components.NiceIntegerSlider;
import com.fluxicon.slickerbox.components.NiceSlider.Orientation;
import com.fluxicon.slickerbox.factory.SlickerFactory;

/**
 * @author F. Mannhardt
 *
 */
public class BalancedAlignmentConfigPanel extends JPanel {
	
	private static final long serialVersionUID = 4999650753382137345L;
	
	private static final int MAXLIMMAXNUMINSTANCES = 100_000_001;
	private static final int DEFLIMMAXNUMINSTANCES = 100_000_001;
	
	private ProMPropertiesExpertModePanel panel;

	private NiceIntegerSlider concurrentThreads;
	
	private JCheckBox dataCacheActivated;
	private JCheckBox useOptimizations;
	private JCheckBox keepControlFlowSearchSpace;
	private JCheckBox keepDataFlowSearchSpace;
	private JCheckBox usePartialDataAlignments;
	
	private ProMComboBox<SearchMethod> searchMethod;

	private ProMComboBox<ASynchronousMoveSorting> sorting;
	private ProMComboBox<QueueingModel> queueingModel;
	private ProMComboBox<ILPSolver> ilpSolver;
	private ProMComboBox<UnassignedMode> unassignedMode;

	private NiceDoubleSlider maxCostInQueue;
	private NiceIntegerSlider limitQueuedStates;

	private JCheckBox usePartialOrder;
	
	public BalancedAlignmentConfigPanel() {
		setPreferredSize(new Dimension(600, 800));
		setBackground(WidgetColors.PROPERTIES_BACKGROUND);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		
		JLabel introduction = new JLabel("<html>"
				+ "You are about to compute an alignment a Data Petri Net (DPN) (or plain Petri net) and an Event Log. "
				+ "Please note that a more user-friendly (and up-to-date) alternative to this plug-ins is the: <u>Multi-perspective Process Explorer</u>. "
				+ "For most parameters using the <u>default option</u> is a reasonable choice."
				+ "<ul>"
				+ "<li>Performance and Algorithm-related Parameters (this dialog: use defaults)</li>"
				+ "<li>Mapping between Events and Transitions (<ub>important</u>: check correct mapping)</li>"
				+ "<li>Cost for control-flow violations (default cost is 1)</li>"
				+ "<li>Mapping between variables in the DPN and attributes in the event log (<u>important</u>: check correct mapping)</li>"				
				+ "<li>Cost for data violations (default cost is 1)</li>"
				+ "<li>Bounds for variables (use defaults)</li>"
				+ "</ul>"
				+ "The computation is computationally very intensive depending on the used model (degree of parallelism, complexity of guards) and log (length of traces). "
				+ "It is advised to have <b>at least 4GB of memory</b> allocated to ProM. Reducing the number of concurrent threads might help to reduce memory requirements. "
				+ "</html>");
		introduction.setAlignmentX(LEFT_ALIGNMENT);
		add(introduction);
		
		panel = new ProMPropertiesExpertModePanel(null);
		panel.setPreferredSize(null);
		panel.setMaximumSize(null);
		panel.setMinimumSize(null);
		panel.setBackground(WidgetColors.PROPERTIES_BACKGROUND);
		panel.setAlignmentX(LEFT_ALIGNMENT);
		
		SlickerFactory sF = SlickerFactory.instance();
		
		int availableProcessors = Runtime.getRuntime().availableProcessors();
		concurrentThreads = sF.createNiceIntegerSlider("", 1, availableProcessors, availableProcessors, Orientation.HORIZONTAL);
		panel.addProperty("Concurrent threads" , concurrentThreads);
		panel.addProperty("", createLabel("How much concurrent threads to use (lower value uses less memory)."), false);
		
		maxCostInQueue = sF.createNiceDoubleSlider("", 0, 1, 0, Orientation.HORIZONTAL);
		panel.addProperty("Limit fitness", maxCostInQueue, false);
		panel.addProperty("", createLabel("Stops the computation, if fitness for trace would drop below given value."), false);
		
		limitQueuedStates = sF.createNiceIntegerSlider("", 1, MAXLIMMAXNUMINSTANCES, DEFLIMMAXNUMINSTANCES, Orientation.HORIZONTAL);
		panel.addProperty("Limit search space", limitQueuedStates, false);
		panel.addProperty("", createLabel("Stops the computation, if the given number of states has been queued. "+MAXLIMMAXNUMINSTANCES+" means unlimited!"), false);
		
		unassignedMode = panel.addComboBox("Unassigned variable mode", UnassignedMode.values(), false);
		unassignedMode.setSelectedItem(UnassignedMode.NULL);
		panel.addProperty("", createLabel("Determines how unassigned variables are handled."), false);
		
		ilpSolver = panel.addComboBox("MILP Solver", DataAlignmentAdapter.ILPSolver.values(), false);
		ilpSolver.setSelectedItem(DataAlignmentAdapter.ILPSolver.ILP_LPSOLVE);
		panel.addProperty("", createLabel("Solver to be used for MILP problems. Gurobi is faster!"), false);
		
		searchMethod = panel.addComboBox("Search Algorithm", SearchMethod.values(), true);
		searchMethod.setSelectedItem(SearchMethod.ASTAR_GRAPH);
		panel.addProperty("", createLabel("Algorithm that is used to find an (optimal) alignment. A* graph is fastest."), true);
		
		usePartialDataAlignments = panel.addCheckBox("Compute Balanced Alignment", true, true);
		panel.addProperty("", createLabel("Use a balanced alignment. (Disable to use staged approach of BPM 2013 paper"), true);		

		keepControlFlowSearchSpace = panel.addCheckBox("Keep Control Flow Search Space", true, true);
		panel.addProperty("", createLabel("Retain the search space between traces to speed up the computation. (Disable to use less memory)"), true);
		
		keepDataFlowSearchSpace = panel.addCheckBox("Keep Data Flow Search Space", true, true);
		panel.addProperty("", createLabel("Retain the search space between traces to speed up the computation. (Disable to use less memory)"), true);
		
		dataCacheActivated = panel.addCheckBox("Use Cache", true, true);
		panel.addProperty("", createLabel("Remember the solution to MILP problems and re-uses those solutions if possible. (Disable to use less memory)"), true);
		
		useOptimizations = panel.addCheckBox("Use Optimization", true, true);
		panel.addProperty("", createLabel("Activates optimizations that avoid solving a ILP problem. (Disable to be slower)"), true);			

		usePartialOrder = panel.addCheckBox("Use PartialOrder", false, true);
		panel.addProperty("", createLabel("Interprets events with identical timestamps as partially ordered (Enabled is slower)"), true);			
		
		sorting = panel.addComboBox("Re-Ordering of Moves", ASynchronousMoveSorting.values(), true);
		sorting.setSelectedItem(ASynchronousMoveSorting.LOGMOVEFIRST);
		panel.addProperty("", createLabel("Whether to search all re-orderings (NONE) or just a single special ordering between log-moves and model-moves."), true);

		// Only for plain A*
		queueingModel = panel.addComboBox("Queueing Strategy", QueueingModel.values(), true);
		queueingModel.setSelectedItem(QueueingModel.DEPTHFIRST);
		panel.addProperty("", createLabel("Strategy to break ties in the A* search. Random can be useful to get multiple solutions with equal cost."), true);
		
		add(panel);
	}

	private JComponent createLabel(String s) {
		JLabel label = new JLabel("<HTML>"+s+"</HTML>");
		label.setForeground(Color.LIGHT_GRAY);
		return label;
	}

	public boolean isDataViewCacheActivated() {
		return dataCacheActivated.getModel().isSelected();
	}

	public boolean getIsUseOptimizations() {
		return useOptimizations.getModel().isSelected();
	}

	public boolean getIsUsePartialOrders() {
		return usePartialOrder.getModel().isSelected();
	}

	public int getConcurrentThreads() {
		return concurrentThreads.getValue();
	}

	public boolean getUseILP() {
		return true;
	}

	public ASynchronousMoveSorting getSorting() {
		return (ASynchronousMoveSorting) sorting.getSelectedItem();
	}
	
	public DataAlignmentAdapter.ILPSolver getILPSolver() {
		return (DataAlignmentAdapter.ILPSolver) ilpSolver.getSelectedItem();
	}

	public QueueingModel getQueueingModel() {
		return (QueueingModel) queueingModel.getSelectedItem();
	}

	public double getMaxCost() {
		return maxCostInQueue.getValue();
	}

	public SearchMethod getSearchMethod() {
		return (SearchMethod) searchMethod.getSelectedItem();
	}
	
	public int getMaxQueuedStates() {
		return limitQueuedStates.getValue() == MAXLIMMAXNUMINSTANCES ? Integer.MAX_VALUE : limitQueuedStates.getValue();
	}
	
	public boolean getIsUsePartialDataAlignments() {
		return usePartialDataAlignments.getModel().isSelected();
	}
	
	public UnassignedMode getUnassignedMode() {
		return (UnassignedMode) unassignedMode.getSelectedItem();
	}

	public boolean getIsKeepControlFlowSearchSpace() {
		return keepControlFlowSearchSpace.getModel().isSelected();
	}

	public boolean getIsKeepDataFlowSearchSpace() {
		return keepDataFlowSearchSpace.getModel().isSelected();
	}

}
