/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.visualisation;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Paint;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.NumberFormat;
import java.util.Locale;

import javax.swing.JComponent;
import javax.swing.JFileChooser;

import org.apache.commons.math3.stat.StatUtils;
import org.freehep.graphicsio.ps.EPSGraphics2D;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.StandardChartTheme;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.PaintScale;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYDotRenderer;
import org.jfree.chart.title.PaintScaleLegend;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.RectangleInsets;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.WidgetColors;
import org.processmining.plugins.balancedconformance.result.BalancedReplayResult;

import com.fluxicon.slickerbox.components.SlickerTabbedPane;
import com.fluxicon.slickerbox.factory.SlickerFactory;

/**
 * Some detailed statistics about the computation.
 * 
 * @author F. Mannhardt
 *
 */
@Plugin(name = "9 Statistics about the computation", level = PluginLevel.Local, //
		returnLabels = { "Balanced Data Conformance Checking Statistics" }, returnTypes = { JComponent.class }, // 
		userAccessible = true, parameterLabels = { "Matching Instances" })
@Visualizer
public class BalancedDataConformanceStatsVis {

	private static final class ChartListener implements MouseListener {
		private final double[] dataY;
		private final double[] dataX;
		private final ChartPanel panel;

		private ChartListener(double[] dataY, double[] dataX, ChartPanel panel) {
			this.dataY = dataY;
			this.dataX = dataX;
			this.panel = panel;
		}

		private ChartListener(double[] dataX, ChartPanel panel) {
			this.dataY = null;
			this.dataX = dataX;
			this.panel = panel;
		}

		public void mouseReleased(MouseEvent e) {
		}

		public void mousePressed(MouseEvent e) {
		}

		public void mouseExited(MouseEvent e) {
		}

		public void mouseEntered(MouseEvent e) {
		}

		public void mouseClicked(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON2) {
				final JFileChooser fc = new JFileChooser();
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int retVal = fc.showSaveDialog(panel);
				if (retVal == JFileChooser.APPROVE_OPTION) {
					exportChart(fc.getSelectedFile(), panel, dataX, dataY);
				}
			}
		}

		private void exportChart(File file, ChartPanel panel, double[] dataX, double[] dataY) {
			export(new File(file.getAbsolutePath() + ".eps"), panel.getChart(), 400, 400);
			exportCSV(new File(file.getAbsolutePath() + ".csv"), dataX, dataY);
		}

		private void export(File file, JFreeChart chart, int x, int y) {
			try {
				EPSGraphics2D epsGraphics = new EPSGraphics2D(file, new Dimension(x, y));
				chart.draw(epsGraphics, new Rectangle(x, y));
				epsGraphics.closeStream();
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		protected void exportCSV(File selectedFile, double[] dataX, double[] dataY) {
			BufferedWriter bw = null;
			try {

				bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(selectedFile)));
				bw.write("\"dataX\"");
				if (dataY != null) {
					bw.write(",\"dataY\"");
				}

				final NumberFormat fNF = NumberFormat.getInstance(Locale.ENGLISH);
				fNF.setMaximumFractionDigits(5);
				fNF.setMinimumFractionDigits(0);

				for (int j = 0; j < dataX.length; j++) {
					bw.newLine();
					bw.write("\"" + fNF.format(dataX[j]) + "\"");
					if (dataY != null && j < dataY.length) {
						bw.write(",\"" + fNF.format(dataY[j]) + "\"");
					}
				}

			} catch (FileNotFoundException e) {
				throw new RuntimeException(e);
			} catch (IOException e) {
				throw new RuntimeException(e);
			} finally {
				if (bw != null) {
					try {
						bw.close();
					} catch (Exception exc) {
						// do nothing
					}
				}
			}
		}
	}

	private static final class MyPaintScale implements PaintScale {

		private double lowerBound;
		private double upperBound;

		public MyPaintScale(double min, double max) {
			super();
			this.lowerBound = min;
			this.upperBound = max;
		}

		public double getLowerBound() {
			return lowerBound;
		}

		public double getUpperBound() {
			return upperBound;
		}

		public Paint getPaint(double value) {
			double v = Math.max(value, this.lowerBound);
			v = Math.min(v, this.upperBound);
			int g = 240 - (int) ((value - this.lowerBound) / (this.upperBound - this.lowerBound) * 240.0);
			return new Color(g, g, g);
		}

	}

	private static final class ColorXYItemRenderer extends XYDotRenderer {

		private static final long serialVersionUID = 8847883784765968521L;

		private final double[] dataZ;
		private final PaintScale paintScale;

		public ColorXYItemRenderer(double[] dataZ) {
			super();
			this.setDotHeight(5);
			this.setDotWidth(10);
			this.paintScale = new MyPaintScale(StatUtils.min(dataZ), StatUtils.max(dataZ));
			this.dataZ = dataZ;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.jfree.chart.renderer.AbstractRenderer#getItemPaint(int, int)
		 */
		@Override
		public Paint getItemPaint(int row, int column) {
			return paintScale.getPaint(dataZ[column]);
		}

		public PaintScale getPaintScale() {
			return paintScale;
		}

	}

	private static final class StandardXYItemRenderer extends XYDotRenderer {
		private static final long serialVersionUID = 1018391957000784888L;

		public StandardXYItemRenderer() {
			super();
			this.setDotHeight(5);
			this.setDotWidth(5);
			setSeriesPaint(0, Color.BLACK);
			setSeriesPaint(1, Color.ORANGE);
		}
	}

	@PluginVariant(requiredParameterLabels = { 0 })
	public JComponent visualise(PluginContext context, BalancedReplayResult res) {

		initializeChartDefaults();

		StatisticsPanelBuilder statsBuilder = new StatisticsPanelBuilder();
		ProMSplitPane parentPanel = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		JComponent statsPanel = statsBuilder.createMetaDataPanelWithoutButtons(context, res);
		statsPanel.setPreferredSize(new Dimension(800, 400));
		parentPanel.setLeftComponent(statsPanel);

		SlickerTabbedPane chartPane = SlickerFactory.instance().createTabbedPane("", WidgetColors.COLOR_LIST_BG,
				WidgetColors.COLOR_LIST_FG, Color.WHITE);

		double[] dataTime = res.getStatisticsStore().get("timePerTrace").getData();
		double[] dataFitness = res.getStatisticsStore().get("fitness").getData();
		double[] dataLength = res.getStatisticsStore().get("traceLength").getData();
		double[] dataCacheHits = res.getStatisticsStore().get("cacheHits").getData();
		double[] dataCacheSize = res.getStatisticsStore().get("cacheSize").getData();

		try {
			chartPane.addTab("Time vs. Fitness",
					addChart(dataFitness, dataTime, "Time vs. Fitness", "Fitness", "Time (ms)", true, false));
			chartPane.addTab("Time vs. Length",
					addChart(dataLength, dataTime, "Time vs. Length", "Length", "Time (ms)", true, false));
			chartPane.addTab("Time vs. Length (Grayscaled)", addHeatChart(dataLength, dataTime, dataFitness,
					"Time vs. Length (Grayscaled)", "Length", "Time (ms)", true, false));
			chartPane.addTab("Cache Size vs. Cache Hits", addChart(dataCacheSize, dataCacheHits,
					"Cache Size vs. Cache Hits", "Cache Size", "Cache Hits", false, true));
		} catch (IllegalArgumentException e) {
		}
		parentPanel.setRightComponent(chartPane);

		return parentPanel;
	}

	private void initializeChartDefaults() {
		StandardChartTheme theme = (StandardChartTheme) org.jfree.chart.StandardChartTheme.createJFreeTheme();
		theme.setChartBackgroundPaint(Color.WHITE);
		theme.setShadowVisible(false);
		theme.setPlotBackgroundPaint(Color.WHITE);
		theme.setRangeGridlinePaint(Color.LIGHT_GRAY);
		theme.setDomainGridlinePaint(Color.LIGHT_GRAY);
		theme.setAxisLabelPaint(Color.BLACK);
		theme.setRegularFont(new Font("SansSerif", Font.PLAIN, 12));
		theme.setLargeFont(new Font("SansSerif", Font.PLAIN, 14));
		theme.setExtraLargeFont(new Font("SansSerif", Font.PLAIN, 16));
		theme.setSmallFont(new Font("SansSerif", Font.PLAIN, 10));
		ChartFactory.setChartTheme(theme);
		XYBarRenderer.setDefaultShadowsVisible(false);
	}

	private ChartPanel addChart(double[] dataX, double[] dataY, String chartName, String xAxis, String yAxis,
			boolean showLogarithmicScale, boolean allowZero) {
		assert dataX.length == dataY.length : "Data X and Data Y does not match in size";
		XYSeries series = new XYSeries("default", false, true);
		series.setNotify(false);
		for (int i = 0; i < dataX.length; i++) {
			if (dataY[i] > 0 || allowZero) {
				series.add(dataX[i], dataY[i]);
			}
		}
		series.setNotify(true);
		XYDataset dataset = new XYSeriesCollection(series);
		JFreeChart chart = ChartFactory.createScatterPlot(chartName, xAxis, yAxis, dataset, PlotOrientation.VERTICAL,
				false, false, false);
		XYPlot plot = chart.getXYPlot();
		plot.setRenderer(new StandardXYItemRenderer());
		if (showLogarithmicScale) {
			LogarithmicAxis logarithmicAxis = new LogarithmicAxis(yAxis);
			logarithmicAxis.setStrictValuesFlag(false);
			plot.setRangeAxis(logarithmicAxis);
		}
		ChartPanel panel = new ChartPanel(chart, false);
		panel.addMouseListener(new ChartListener(dataY, dataX, panel));
		return panel;
	}

	private ChartPanel addHeatChart(double[] dataX, double[] dataY, double[] dataZ, String chartName, String xAxis,
			String yAxis, boolean showLogarithmicScale, boolean allowZero) {
		assert dataX.length == dataY.length : "Data X and Data Y does not match in size";
		XYSeries series = new XYSeries(chartName, false, true);
		series.setNotify(false);
		for (int i = 0; i < dataX.length; i++) {
			if (dataY[i] > 0 || allowZero) {
				series.add(dataX[i], dataY[i]);
			}
		}
		series.setNotify(true);
		XYDataset dataset = new XYSeriesCollection(series);
		JFreeChart chart = ChartFactory.createScatterPlot("", xAxis, yAxis, dataset, PlotOrientation.VERTICAL, true,
				false, false);
		XYPlot plot = chart.getXYPlot();
		plot.setBackgroundPaint(Color.WHITE);
		plot.setDomainMinorGridlinePaint(Color.LIGHT_GRAY);
		plot.setDomainGridlinePaint(Color.LIGHT_GRAY);
		plot.setRangeMinorGridlinePaint(Color.LIGHT_GRAY);
		plot.setRangeGridlinePaint(Color.LIGHT_GRAY);
		plot.getRangeAxis().setLabelFont(new Font("SansSerif", Font.PLAIN, 12));
		plot.getDomainAxis().setLabelFont(new Font("SansSerif", Font.PLAIN, 12));
		ColorXYItemRenderer renderer = new ColorXYItemRenderer(dataZ);
		plot.setRenderer(renderer);
		if (showLogarithmicScale) {
			LogarithmicAxis logarithmicAxis = new LogarithmicAxis(yAxis);
			logarithmicAxis.setStrictValuesFlag(false);
			plot.setRangeAxis(logarithmicAxis);
		}

		chart.removeLegend();
		PaintScaleLegend legend = new PaintScaleLegend(renderer.getPaintScale(), new NumberAxis(""));
		legend.setMargin(new RectangleInsets(0, 0, 5, 0));
		legend.setPadding(new RectangleInsets(0, 40, 0, 10));
		legend.setStripWidth(10);
		legend.setPosition(RectangleEdge.BOTTOM);
		chart.addSubtitle(legend);
		ChartPanel panel = new ChartPanel(chart, false);
		panel.addMouseListener(new ChartListener(dataY, dataX, panel));
		return panel;
	}

}
