/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.visualisation;

import java.awt.Dimension;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.NumberFormat;
import java.util.Locale;

import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.table.DefaultTableModel;

import org.apache.commons.lang3.math.NumberUtils;
import org.apache.commons.math3.stat.StatUtils;
import org.apache.commons.math3.stat.descriptive.moment.StandardDeviation;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.StandardChartTheme;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.data.statistics.HistogramDataset;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.plugins.balancedconformance.result.BalancedReplayResult;
import org.processmining.plugins.balancedconformance.result.StatisticResult;

/**
 * @author F. Mannhardt
 *
 */
class StatisticsPanelBuilder {

	private StatisticResult[] statistics;
	
	public JComponent createMetaDataPanelWithoutButtons(final PluginContext context,
			final BalancedReplayResult logReplayResult) {

		final NumberFormat intNF = NumberFormat.getInstance();
		intNF.setMaximumFractionDigits(0);
		intNF.setMinimumFractionDigits(0);

		statistics = logReplayResult.getStatisticsStore().values().toArray(new StatisticResult[0]);		
		
		String[][] metaData = new String[statistics.length + 2][];
		int i = 0;
		for (StatisticResult res : statistics) {
			if (res.isBoolean()) {
				metaData[i++] = calcStats(res.getName(), res.getDataBool());
			} else {
				metaData[i++] = calcStats(res.getName(), res.getData());
			}
		}		
		metaData[i] = new String[] { "Calculation Time", intNF.format(logReplayResult.getCalcTime()), "", "", "", "",
				"", "", "" };

		DefaultTableModel tableModel = new DefaultTableModel(metaData, new String[] { "Property", "Count", "Min",
				"Avg", "Q1", "Med", "Q3", "Max", "Std" }) {

			private static final long serialVersionUID = 5254787976189031958L;

			@Override
			public boolean isCellEditable(int row, int column) {
				return false;
			}

		};

		final ProMTable promTable = new ProMTable(tableModel);
		promTable.getTable().addMouseListener(new MouseListener() {

			public void mouseReleased(MouseEvent e) {
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseExited(MouseEvent e) {
			}

			public void mouseEntered(MouseEvent e) {
			}

			public void mouseClicked(MouseEvent e) {
				if (e.getClickCount() >= 2) {
					int rowIndex = promTable.rowAtPoint(e.getPoint());
					if (rowIndex < statistics.length) {
						StatisticResult result = statistics[rowIndex];
						String binSize = (String) JOptionPane.showInputDialog(promTable, "Choose bin size:\n",
								"Choose bin size", JOptionPane.QUESTION_MESSAGE, null, null,
								Integer.toString(calcBinSize(result.getData())));
						if (NumberUtils.isNumber(binSize)) {
							showHistogram(result, Integer.parseInt(binSize));	
						} else {
							showHistogram(result);	
						}						
					}
				}
				
				if (e.getButton() == MouseEvent.BUTTON2) {
					int rowIndex = promTable.rowAtPoint(e.getPoint());
					if (rowIndex < statistics.length) {
						StatisticResult result = statistics[rowIndex];
						final JFileChooser fc = new JFileChooser();
						fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
						int retVal = fc.showSaveDialog(promTable);
						if (retVal == JFileChooser.APPROVE_OPTION) {
							exportCSV(context, result, fc.getSelectedFile());
						}
					}
				}
			}

		});

		return promTable;

	}

	protected void exportCSV(PluginContext context, StatisticResult result, File selectedFile) {
			BufferedWriter bw = null;
			try {

				bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(selectedFile)));
				bw.write("\""+result.getName()+"\"");
				
				final NumberFormat fNF = NumberFormat.getInstance(Locale.ENGLISH);
				fNF.setMaximumFractionDigits(5);
				fNF.setMinimumFractionDigits(0);
				
				for (int j = 0; j < result.getData().length; j++) {
					bw.newLine();
					bw.write("\""+fNF.format(result.getData()[j])+"\"");
				}
				
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} finally {
				if (bw != null) {
					try {
						bw.close();
					} catch (Exception exc) {
						// do nothing
					}
				}
			}
	}

	public JComponent createMetaDataPanel(final PluginContext context,
			final BalancedReplayResult logReplayResult) {
		return createMetaDataPanelWithoutButtons(context, logReplayResult);
	}

	private static String[] calcStats(String caption, double[] doubleArray) {
		return new String[] { caption, sum(doubleArray), min(doubleArray), average(doubleArray),
				firstQuartile(doubleArray), secondQuartile(doubleArray), thirdQuartile(doubleArray), max(doubleArray),
				stdDev(doubleArray) };
	}

	private static String[] calcStats(String caption, boolean[] boolArray) {
		final NumberFormat intNF = NumberFormat.getInstance();
		intNF.setMaximumFractionDigits(0);
		intNF.setMinimumFractionDigits(0);
		return new String[] { caption, intNF.format(sumBoolean(boolArray)), "", "", "", "", "", "", "" };
	}

	public static int sumBoolean(boolean[] boolArray) {
		int count = 0;
		for (int i = 0; i < boolArray.length; i++) {
			if (boolArray[i]) {
				count++;
			}
		}
		return count;
	}

	private static String sum(double[] data) {
		return formatDoubleAsString(StatUtils.sum(data));
	}

	private static String max(double[] data) {
		return formatDoubleAsString(StatUtils.max(data));
	}

	private static String formatDoubleAsString(double d) {
		final NumberFormat floatNF = NumberFormat.getInstance(Locale.GERMAN);
		floatNF.setMaximumFractionDigits(2);
		floatNF.setMinimumFractionDigits(0);
		return floatNF.format(d);
	}

	private static String thirdQuartile(double[] data) {
		return formatDoubleAsString(StatUtils.percentile(data, 75.0));
	}

	private static String secondQuartile(double[] data) {
		return formatDoubleAsString(StatUtils.percentile(data, 50.0));
	}

	private static String firstQuartile(double[] data) {
		return formatDoubleAsString(StatUtils.percentile(data, 25.0));
	}

	private static String average(double[] data) {
		return formatDoubleAsString(StatUtils.mean(data));
	}

	private static String min(double[] data) {
		return formatDoubleAsString(StatUtils.min(data));
	}

	private static String stdDev(double[] data) {
		return formatDoubleAsString(new StandardDeviation(false).evaluate(data));
	}

	protected void showHistogram(StatisticResult result) {
		if (!result.isBoolean() && result.getxAxis() != null) {
			createHistogram(result.getName(), result.getxAxis(), result.getyAxis(), result.getData());
		}
	}
	
	protected void showHistogram(StatisticResult result, int binSize) {
		if (!result.isBoolean() && result.getxAxis() != null) {
			createHistogram(result.getName(), result.getxAxis(), result.getyAxis(), result.getData(), binSize);
		}
	}

	private static void createHistogram(String title, String xAxisTitle, String yAxisTitle, double[] data) {
		int binSize = calcBinSize(data);
		createHistogram(title, xAxisTitle, yAxisTitle, data, binSize);
	}

	private static int calcBinSize(double[] data) {
		double binWidth = Math.max(0.01,
				(2 * (StatUtils.percentile(data, 75.0) - StatUtils.percentile(data, 25.0)) * Math.pow(data.length,
						-(1.0 / 3.0))));
		int binSize = Math.max(10,
				(int) Math.ceil((StatUtils.percentile(data, 99.0) - StatUtils.percentile(data, 1.0)) / binWidth));
		return binSize;
	}

	private static void createHistogram(String title, String xAxisTitle, String yAxisTitle, double[] data, int binSize) {
		if (data.length > 0) {
			HistogramDataset dataset = new HistogramDataset();
			dataset.addSeries("Main", data, binSize);
			ChartFactory.setChartTheme(StandardChartTheme.createLegacyTheme());
			JFreeChart chart = ChartFactory.createHistogram(title, xAxisTitle, yAxisTitle, dataset,
					PlotOrientation.VERTICAL, false, true, false);
			XYPlot plot = chart.getXYPlot();
			XYBarRenderer renderer = (XYBarRenderer) plot.getRenderer();
			renderer.setShadowVisible(false);
			ChartPanel panel = new ChartPanel(chart, false);
			panel.setPreferredSize(new Dimension(640, 480));
			JFrame frame = new JFrame(title);
			frame.add(panel);
			frame.pack();
			frame.setVisible(true);
		}
	}

}
